"""Formal runtime invariant checks for Orchesis."""

from __future__ import annotations

import random
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path

from orchesis.config import load_policy
from orchesis.drift import DriftDetector
from orchesis.engine import RULE_EVALUATION_ORDER, evaluate
from orchesis.identity import AgentIdentity, AgentRegistry, TrustTier
from orchesis.replay import ReplayEngine, read_events_from_jsonl
from orchesis.state import RateLimitTracker


@dataclass
class InvariantResult:
    name: str
    passed: bool
    detail: str
    checked_at: str


@dataclass
class InvariantReport:
    results: list[InvariantResult]
    all_passed: bool
    duration_seconds: float


def _now_iso() -> str:
    return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")


class InvariantChecker:
    """Verifies formal runtime invariants."""

    def __init__(self, policy_path: str, decisions_log: str = ".orchesis/decisions.jsonl"):
        self._policy_path = policy_path
        self._decisions_log = Path(decisions_log)
        self._policy = load_policy(policy_path)

    def check_all(self) -> InvariantReport:
        started = time.perf_counter()
        checks = [
            self.check_never_fail_open,
            self.check_deterministic_replay,
            self.check_state_isolation,
            self.check_no_cross_session_leak,
            self.check_fail_closed_on_error,
            self.check_evaluation_order_stable,
            self.check_identity_enforcement,
            self.check_cost_never_negative_allow,
            self.check_rate_limit_atomic,
            self.check_no_state_drift,
        ]
        results = [check() for check in checks]
        return InvariantReport(
            results=results,
            all_passed=all(item.passed for item in results),
            duration_seconds=max(0.0, time.perf_counter() - started),
        )

    def check_never_fail_open(self) -> InvariantResult:
        dangerous_requests = [
            {"tool": "read_file", "params": {"path": "/etc/passwd"}, "cost": 0.1},
            {"tool": "run_sql", "params": {"query": "DROP TABLE users"}, "cost": 0.1},
            {"tool": "api_call", "params": {}, "cost": -1.0},
        ]
        failed = 0
        for request in dangerous_requests:
            decision = evaluate(request, self._policy)
            if decision.allowed:
                failed += 1
        passed = failed == 0
        detail = "dangerous probes denied" if passed else f"{failed} dangerous probes were allowed"
        return InvariantResult("never_fail_open", passed, detail, _now_iso())

    def check_deterministic_replay(self) -> InvariantResult:
        events = read_events_from_jsonl(self._decisions_log)
        if not events:
            return InvariantResult("deterministic_replay", True, "no events to replay", _now_iso())
        rng = random.Random(42)
        sample_size = min(100, len(events))
        sampled = [events[index] for index in rng.sample(range(len(events)), sample_size)]
        replay = ReplayEngine()
        mismatches = 0
        for event in sampled:
            result = replay.replay_event(event, self._policy, strict=False)
            if not result.match:
                mismatches += 1
        passed = mismatches == 0
        detail = "sampled replay matches" if passed else f"{mismatches} replay mismatches"
        return InvariantResult("deterministic_replay", passed, detail, _now_iso())

    def check_state_isolation(self) -> InvariantResult:
        tracker = RateLimitTracker(persist_path=None)
        policy = {"rules": [{"name": "rate_limit", "max_requests_per_minute": 2}]}
        request_a = {
            "tool": "read_file",
            "params": {"path": "/data/a.txt"},
            "context": {"agent": "agent_a"},
        }
        request_b = {
            "tool": "read_file",
            "params": {"path": "/data/b.txt"},
            "context": {"agent": "agent_b"},
        }
        _ = evaluate(request_a, policy, state=tracker)
        _ = evaluate(request_a, policy, state=tracker)
        denied_a = evaluate(request_a, policy, state=tracker)
        allow_b = evaluate(request_b, policy, state=tracker)
        passed = (not denied_a.allowed) and allow_b.allowed
        detail = "per-agent rate isolation OK" if passed else "cross-agent contamination detected"
        return InvariantResult("state_isolation", passed, detail, _now_iso())

    def check_no_cross_session_leak(self) -> InvariantResult:
        tracker = RateLimitTracker(persist_path=None)
        policy = {"rules": [{"name": "rate_limit", "max_requests_per_minute": 2}]}
        s1 = {
            "tool": "read_file",
            "params": {"path": "/data/a.txt"},
            "context": {"agent": "agent", "session": "s1"},
        }
        s2 = {
            "tool": "read_file",
            "params": {"path": "/data/a.txt"},
            "context": {"agent": "agent", "session": "s2"},
        }
        _ = evaluate(s1, policy, state=tracker)
        _ = evaluate(s1, policy, state=tracker)
        deny_s1 = evaluate(s1, policy, state=tracker)
        allow_s2 = evaluate(s2, policy, state=tracker)
        passed = (not deny_s1.allowed) and allow_s2.allowed
        detail = "session isolation OK" if passed else "session leak detected"
        return InvariantResult("no_cross_session_leak", passed, detail, _now_iso())

    def check_fail_closed_on_error(self) -> InvariantResult:
        class BrokenTracker(RateLimitTracker):
            def check_and_record(self, *args, **kwargs):  # noqa: ANN002, ANN003
                raise RuntimeError("broken state")

        tracker = BrokenTracker(persist_path=None)
        policy = {"rules": [{"name": "rate_limit", "max_requests_per_minute": 1}]}
        decision = evaluate(
            {"tool": "read_file", "params": {"path": "/data/a"}}, policy, state=tracker
        )
        passed = not decision.allowed
        detail = "state error denied" if passed else "state error allowed request"
        return InvariantResult("fail_closed_on_error", passed, detail, _now_iso())

    def check_evaluation_order_stable(self) -> InvariantResult:
        policy = {
            "rules": [
                {"name": "budget_limit", "max_cost_per_call": 10.0},
                {"name": "rate_limit", "max_requests_per_minute": 100},
                {"name": "file_access", "denied_paths": ["/etc"]},
                {"name": "sql_restriction", "denied_operations": ["DROP"]},
                {
                    "name": "regex",
                    "type": "regex_match",
                    "field": "params.query",
                    "deny_patterns": ["DROP"],
                },
                {
                    "name": "ctx",
                    "type": "context_rules",
                    "rules": [{"agent": "*", "max_cost_per_call": 100.0}],
                },
                {
                    "name": "combo",
                    "type": "composite",
                    "operator": "AND",
                    "conditions": [{"rule": "budget_limit"}],
                },
            ]
        }
        request = {
            "tool": "run_sql",
            "params": {"query": "SELECT 1", "path": "/data/x.txt"},
            "cost": 0.1,
            "context": {"agent": "cursor"},
        }
        decision = evaluate(request, policy, state=RateLimitTracker(persist_path=None))
        expected = [name for name in RULE_EVALUATION_ORDER if name != "identity_check"]
        passed = decision.rules_checked == expected
        detail = "rule order stable" if passed else "rule order mismatch"
        return InvariantResult("evaluation_order_stable", passed, detail, _now_iso())

    def check_identity_enforcement(self) -> InvariantResult:
        registry = AgentRegistry(
            agents={
                "blocked": AgentIdentity(
                    agent_id="blocked", name="Blocked", trust_tier=TrustTier.BLOCKED
                ),
                "intern": AgentIdentity(
                    agent_id="intern", name="Intern", trust_tier=TrustTier.INTERN
                ),
            }
        )
        policy = {"rules": [{"name": "budget_limit", "max_cost_per_call": 10.0}]}
        blocked = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/x.txt"},
                "context": {"agent": "blocked"},
            },
            policy,
            registry=registry,
        )
        intern_read = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/x.txt"},
                "context": {"agent": "intern"},
            },
            policy,
            registry=registry,
        )
        intern_write = evaluate(
            {
                "tool": "write_file",
                "params": {"path": "/data/x.txt"},
                "context": {"agent": "intern"},
            },
            policy,
            registry=registry,
        )
        passed = (not blocked.allowed) and intern_read.allowed and (not intern_write.allowed)
        detail = "identity controls enforced" if passed else "identity control mismatch"
        return InvariantResult("identity_enforcement", passed, detail, _now_iso())

    def check_cost_never_negative_allow(self) -> InvariantResult:
        policy = {"rules": [{"name": "budget_limit", "max_cost_per_call": 1.0}]}
        decision = evaluate(
            {"tool": "read_file", "params": {"path": "/data/safe.txt"}, "cost": -1.0}, policy
        )
        passed = not decision.allowed
        detail = "negative cost denied" if passed else "negative cost allowed"
        return InvariantResult("cost_never_negative_allow", passed, detail, _now_iso())

    def check_rate_limit_atomic(self) -> InvariantResult:
        tracker = RateLimitTracker(persist_path=None)
        policy = {"rules": [{"name": "rate_limit", "max_requests_per_minute": 20}]}
        request = {
            "tool": "read_file",
            "params": {"path": "/data/a.txt"},
            "context": {"agent": "atomic"},
        }

        def _call() -> bool:
            return evaluate(request, policy, state=tracker).allowed

        with ThreadPoolExecutor(max_workers=50) as pool:
            results = list(pool.map(lambda _: _call(), range(100)))
        allowed = sum(1 for item in results if item)
        passed = allowed == 20
        detail = "atomic boundary respected" if passed else f"expected 20 allow, got {allowed}"
        return InvariantResult("rate_limit_atomic", passed, detail, _now_iso())

    def check_no_state_drift(self) -> InvariantResult:
        detector = DriftDetector()
        tracker = RateLimitTracker(persist_path=None)
        events = detector.run_all_checks(
            tracker=tracker,
            policy=self._policy,
            decisions_log=self._decisions_log,
        )
        passed = not detector.has_critical_drift
        detail = "no critical drift events" if passed else f"{len(events)} drift events detected"
        return InvariantResult("no_state_drift", passed, detail, _now_iso())



"""CLI runner for deterministic agent harness."""

from __future__ import annotations

import argparse
import sys

from orchesis.agent.harness import SimpleAgent


def run_task(task: str, policy_path: str, tasks_path: str, max_steps: int, log_path: str) -> int:
    """Run deterministic agent loop for a given task name."""
    agent = SimpleAgent(
        policy_path=policy_path,
        tools=[],
        tasks_path=tasks_path,
        log_path=log_path,
    )
    state = agent.run(task, max_steps=max_steps)

    print(f"Task: {task}")
    for step in state.steps:
        print(
            f"step={step['step']} tool={step['tool']} decision={step['decision']} "
            f"reasons={step['reasons']}"
        )
    executed = sum(1 for s in state.steps if s["decision"] == "ALLOW")
    blocked = sum(1 for s in state.steps if s["decision"] == "DENY")
    print(f"status={state.status} executed={executed} blocked={blocked}")
    return 0


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="orchesis-agent", description="Orchesis agent harness CLI.")
    subparsers = parser.add_subparsers(dest="command", metavar="command")
    subparsers.required = True
    run_parser = subparsers.add_parser("run")
    run_parser.add_argument("task")
    run_parser.add_argument("--policy", dest="policy_path", required=True)
    run_parser.add_argument("--tasks", dest="tasks_path", default="examples/agent_tasks.yaml")
    run_parser.add_argument("--max-steps", dest="max_steps", type=int, default=10)
    run_parser.add_argument("--log-path", dest="log_path", default="decisions.jsonl")
    return parser


def main(argv: list[str] | None = None) -> int:
    args = build_parser().parse_args(argv)
    if args.command == "run":
        return run_task(
            task=args.task,
            policy_path=args.policy_path,
            tasks_path=args.tasks_path,
            max_steps=args.max_steps,
            log_path=args.log_path,
        )
    return 2


if __name__ == "__main__":
    raise SystemExit(main(sys.argv[1:]))



"""Defensive input validation for text-processing modules."""

from __future__ import annotations


def sanitize_text(text: object) -> str | None:
    """Sanitize text input for downstream scanners/detectors.

    Returns sanitized string, or None if input is not processable text.
    """

    if text is None:
        return None

    if isinstance(text, bytes):
        try:
            text = text.decode("utf-8", errors="replace")
        except Exception:
            return None

    if not isinstance(text, str):
        return None

    if not text or not text.strip():
        return None

    text = text.replace("\x00", "")

    cleaned: list[str] = []
    for ch in text:
        code = ord(ch)
        if code < 32 and ch not in ("\n", "\r", "\t"):
            continue
        if 0xD800 <= code <= 0xDFFF:
            continue
        cleaned.append(ch)
    text = "".join(cleaned)

    if not text.strip():
        return None
    return text


def is_valid_text(text: object) -> bool:
    """Quick check if input is valid processable text.

    Bytes are treated as invalid for this quick-path check.
    """

    if not isinstance(text, str):
        return False
    return sanitize_text(text) is not None




"""Tests for Context Engine — intelligent context window management."""

from __future__ import annotations

import threading
from typing import Any

import pytest

from orchesis.context_engine import (
    ContextConfig,
    ContextEngine,
    ContextResult,
)


# --- Dedup Strategy (8 tests) ---


def test_dedup_consecutive_identical() -> None:
    """Two same messages → one."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "hello"},
        {"role": "user", "content": "hello"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 1
    assert result.messages[0]["content"] == "hello"
    assert "dedup" in result.strategies_applied


def test_dedup_non_consecutive_no_removal() -> None:
    """Same message far apart (beyond dedup_window) → both kept."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"], dedup_window=2)
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "hello"},
        {"role": "assistant", "content": "m1"},
        {"role": "user", "content": "m2"},
        {"role": "assistant", "content": "m3"},
        {"role": "user", "content": "hello"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 5


def test_dedup_preserves_system() -> None:
    """System messages never deduped."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "You are helpful"},
        {"role": "system", "content": "You are helpful"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 2
    assert result.messages[0]["role"] == "system"
    assert result.messages[1]["role"] == "system"


def test_dedup_different_roles_kept() -> None:
    """Same content, different role → both kept."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "hello"},
        {"role": "assistant", "content": "hello"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 2
    assert result.messages[0]["role"] == "user"
    assert result.messages[1]["role"] == "assistant"


def test_dedup_content_list_format() -> None:
    """Anthropic content blocks deduped correctly."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"])
    engine = ContextEngine(cfg)
    block = [{"type": "text", "text": "same"}]
    msgs = [
        {"role": "user", "content": block},
        {"role": "user", "content": block},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 1


def test_dedup_empty_messages() -> None:
    """Empty list → empty list."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"])
    engine = ContextEngine(cfg)
    result = engine.optimize([])
    assert result.messages == []
    assert result.original_count == 0
    assert result.final_count == 0


def test_dedup_window_limit() -> None:
    """Beyond dedup_window → no dedup for far-apart duplicates."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"], dedup_window=2)
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "a"},
        {"role": "user", "content": "b"},
        {"role": "user", "content": "c"},
        {"role": "user", "content": "a"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) >= 3


def test_dedup_single_message() -> None:
    """One message → unchanged."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"])
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": "hello"}]
    result = engine.optimize(msgs)
    assert len(result.messages) == 1
    assert result.messages[0]["content"] == "hello"


def test_dedup_large_window_performance() -> None:
    """Dedup should handle large batches without quadratic slowdowns."""
    import time

    engine = ContextEngine(ContextConfig(enabled=True, strategies=["dedup"], dedup_window=50))
    messages = [{"role": "user", "content": f"msg {i}"} for i in range(1000)]
    for i in range(0, 1000, 7):
        messages[i] = {"role": "user", "content": f"msg {i % 50}"}
    start = time.monotonic()
    result = engine.optimize(messages)
    elapsed = time.monotonic() - start
    assert elapsed < 1.0
    assert len(result.messages) < len(messages)


def test_dedup_window_boundary() -> None:
    """Message outside window should not be considered duplicate."""
    engine = ContextEngine(ContextConfig(enabled=True, strategies=["dedup"], dedup_window=3))
    messages = [
        {"role": "user", "content": "hello"},
        {"role": "user", "content": "world"},
        {"role": "user", "content": "foo"},
        {"role": "user", "content": "bar"},
        {"role": "user", "content": "hello"},
    ]
    result = engine.optimize(messages)
    assert len(result.messages) == 5


def test_dedup_within_window() -> None:
    """Message within window should be deduped."""
    engine = ContextEngine(ContextConfig(enabled=True, strategies=["dedup"], dedup_window=5))
    messages = [
        {"role": "user", "content": "hello"},
        {"role": "user", "content": "world"},
        {"role": "user", "content": "hello"},
    ]
    result = engine.optimize(messages)
    assert len(result.messages) == 2


# --- System Dups Strategy (5 tests) ---


def test_trim_system_dups_identical() -> None:
    """Two identical system → first kept."""
    cfg = ContextConfig(enabled=True, strategies=["trim_system_dups"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "Be helpful"},
        {"role": "system", "content": "Be helpful"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 1
    assert result.messages[0]["content"] == "Be helpful"
    assert "trim_system_dups" in result.strategies_applied


def test_trim_system_dups_different() -> None:
    """Different system messages → both kept."""
    cfg = ContextConfig(enabled=True, strategies=["trim_system_dups"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "Be helpful"},
        {"role": "system", "content": "Be concise"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 2


def test_trim_system_dups_multiple() -> None:
    """3 identical → first kept only."""
    cfg = ContextConfig(enabled=True, strategies=["trim_system_dups"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "X"},
        {"role": "system", "content": "X"},
        {"role": "system", "content": "X"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 1


def test_trim_system_dups_mixed() -> None:
    """System among user/assistant → correct removal."""
    cfg = ContextConfig(enabled=True, strategies=["trim_system_dups"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "S1"},
        {"role": "user", "content": "u1"},
        {"role": "system", "content": "S1"},
        {"role": "assistant", "content": "a1"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 3
    system_contents = [m["content"] for m in result.messages if m["role"] == "system"]
    assert system_contents == ["S1"]


def test_trim_system_dups_none() -> None:
    """No system messages → unchanged."""
    cfg = ContextConfig(enabled=True, strategies=["trim_system_dups"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "hi"},
        {"role": "assistant", "content": "hello"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 2


# --- Tool Result Trimming (10 tests) ---


def test_trim_tool_result_short() -> None:
    """Below threshold → unchanged."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=2000)
    engine = ContextEngine(cfg)
    short = "x" * 100
    msgs = [{"role": "tool", "content": short}]
    result = engine.optimize(msgs)
    assert result.messages[0]["content"] == short
    assert "trim_tool_results" not in result.strategies_applied


def test_trim_tool_result_long() -> None:
    """Above threshold → truncated with marker."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=100)
    engine = ContextEngine(cfg)
    long_content = "x" * 10000
    msgs = [{"role": "tool", "content": long_content}]
    result = engine.optimize(msgs)
    content = result.messages[0]["content"]
    assert len(content) < len(long_content)
    assert "truncated by Orchesis" in content
    assert "trim_tool_results" in result.strategies_applied


def test_trim_tool_result_marker_contains_saved_count() -> None:
    """Marker shows tokens saved."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=50)
    engine = ContextEngine(cfg)
    long_content = "x" * 5000
    msgs = [{"role": "tool", "content": long_content}]
    result = engine.optimize(msgs)
    content = result.messages[0]["content"]
    assert "tokens saved" in content


def test_trim_tool_result_anthropic_format() -> None:
    """Content blocks with tool_result type."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=50)
    engine = ContextEngine(cfg)
    long_text = "y" * 5000
    msgs = [
        {
            "role": "user",
            "content": [{"type": "tool_result", "content": long_text}],
        }
    ]
    result = engine.optimize(msgs)
    blocks = result.messages[0]["content"]
    assert isinstance(blocks, list)
    text = blocks[0].get("content", blocks[0].get("text", ""))
    assert "truncated" in text or len(text) < len(long_text)


def test_trim_tool_result_openai_format() -> None:
    """role=='tool' messages."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=50)
    engine = ContextEngine(cfg)
    long_content = "z" * 5000
    msgs = [{"role": "tool", "content": long_content}]
    result = engine.optimize(msgs)
    assert "truncated" in result.messages[0]["content"]


def test_trim_multiple_tools() -> None:
    """Several tools, only long ones trimmed."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=100)
    engine = ContextEngine(cfg)
    short = "a" * 100
    long_content = "b" * 5000
    msgs = [
        {"role": "tool", "content": short},
        {"role": "tool", "content": long_content},
    ]
    result = engine.optimize(msgs)
    assert result.messages[0]["content"] == short
    assert "truncated" in result.messages[1]["content"]


def test_trim_preserves_non_tool() -> None:
    """User/assistant messages untouched."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=50)
    engine = ContextEngine(cfg)
    long_user = "x" * 10000
    msgs = [{"role": "user", "content": long_user}]
    result = engine.optimize(msgs)
    assert result.messages[0]["content"] == long_user


def test_trim_exact_threshold() -> None:
    """At exactly max_tool_result_tokens → no trim."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=100)
    engine = ContextEngine(cfg)
    exact = "x" * 400
    msgs = [{"role": "tool", "content": exact}]
    result = engine.optimize(msgs)
    assert result.messages[0]["content"] == exact


def test_compress_very_long_output() -> None:
    """4x threshold → head+tail preserved."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["compress_tool_outputs"],
        max_tool_result_tokens=500,
    )
    engine = ContextEngine(cfg)
    head = "HEAD" * 100
    tail = "TAIL" * 100
    mid = "M" * 10000
    long_content = head + mid + tail
    msgs = [{"role": "tool", "content": long_content}]
    result = engine.optimize(msgs)
    content = result.messages[0]["content"]
    assert "HEAD" in content or "middle section removed" in content
    assert "compress_tool_outputs" in result.strategies_applied


def test_compress_preserves_boundaries() -> None:
    """First 500 + last 500 tokens intact."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["compress_tool_outputs"],
        max_tool_result_tokens=500,
    )
    engine = ContextEngine(cfg)
    head = "START_" * 200
    tail = "_END" * 200
    mid = "X" * 15000
    long_content = head + mid + tail
    msgs = [{"role": "tool", "content": long_content}]
    result = engine.optimize(msgs)
    content = result.messages[0]["content"]
    assert "START_" in content
    assert "_END" in content or "middle section" in content


# --- Sliding Window (7 tests) ---


def test_sliding_window_basic() -> None:
    """20 messages, window=10 → 10 kept."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["sliding_window"],
        sliding_window_size=10,
    )
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": f"msg{i}"} for i in range(20)]
    result = engine.optimize(msgs)
    assert len(result.messages) == 10
    assert result.messages[0]["content"] == "msg10"
    assert result.messages[-1]["content"] == "msg19"


def test_sliding_window_preserves_system() -> None:
    """System messages always kept."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["sliding_window"],
        sliding_window_size=3,
        preserve_system=True,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "S"},
        {"role": "user", "content": "u1"},
        {"role": "user", "content": "u2"},
        {"role": "user", "content": "u3"},
        {"role": "user", "content": "u4"},
    ]
    result = engine.optimize(msgs)
    assert any(m["role"] == "system" for m in result.messages)
    system_msgs = [m for m in result.messages if m["role"] == "system"]
    assert len(system_msgs) == 1
    assert system_msgs[0]["content"] == "S"


def test_sliding_window_disabled() -> None:
    """window=0 → no change."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["sliding_window"],
        sliding_window_size=0,
    )
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": f"m{i}"} for i in range(10)]
    result = engine.optimize(msgs)
    assert len(result.messages) == 10


def test_sliding_window_smaller_than_messages() -> None:
    """Fewer messages than window → all kept."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["sliding_window"],
        sliding_window_size=100,
    )
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": f"m{i}"} for i in range(5)]
    result = engine.optimize(msgs)
    assert len(result.messages) == 5


def test_sliding_window_order_preserved() -> None:
    """Output in correct chronological order."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["sliding_window"],
        sliding_window_size=3,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "a"},
        {"role": "assistant", "content": "b"},
        {"role": "user", "content": "c"},
        {"role": "assistant", "content": "d"},
    ]
    result = engine.optimize(msgs)
    assert [m["content"] for m in result.messages] == ["b", "c", "d"]


def test_sliding_window_system_plus_recent() -> None:
    """System + last N-system_count non-system."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["sliding_window"],
        sliding_window_size=4,
        preserve_system=True,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "S"},
        {"role": "user", "content": "u1"},
        {"role": "user", "content": "u2"},
        {"role": "user", "content": "u3"},
        {"role": "user", "content": "u4"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 4
    assert result.messages[0]["role"] == "system"
    contents = [m["content"] for m in result.messages]
    assert "u2" in contents and "u3" in contents and "u4" in contents


def test_sliding_window_all_system() -> None:
    """All system messages → all kept."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["sliding_window"],
        sliding_window_size=2,
        preserve_system=True,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "S1"},
        {"role": "system", "content": "S2"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 2


# --- Token Budget (8 tests) ---


def test_budget_within_limit() -> None:
    """Already fits → unchanged."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["token_budget"],
        max_context_tokens=100000,
        token_budget_reserve=0,
    )
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": "short"}]
    result = engine.optimize(msgs, max_tokens=100000)
    assert len(result.messages) == 1


def test_budget_removes_oldest_first() -> None:
    """Oldest non-system removed first."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["token_budget"],
        max_context_tokens=1000,
        token_budget_reserve=0,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "x" * 200},
        {"role": "assistant", "content": "y" * 200},
        {"role": "user", "content": "z" * 200},
    ]
    result = engine.optimize(msgs, max_tokens=100)
    assert len(result.messages) < 3 or result.final_tokens <= 100


def test_budget_preserves_system() -> None:
    """System messages never removed."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["token_budget"],
        max_context_tokens=50,
        token_budget_reserve=0,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "S" * 100},
        {"role": "user", "content": "u" * 100},
    ]
    result = engine.optimize(msgs, max_tokens=100)
    assert any(m["role"] == "system" for m in result.messages)


def test_budget_preserves_last_user() -> None:
    """Last user message never removed."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["token_budget"],
        max_context_tokens=30,
        token_budget_reserve=0,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "first"},
        {"role": "assistant", "content": "mid"},
        {"role": "user", "content": "last"},
    ]
    result = engine.optimize(msgs, max_tokens=50)
    user_contents = [m["content"] for m in result.messages if m["role"] == "user"]
    assert "last" in user_contents


def test_budget_with_reserve() -> None:
    """Budget accounts for response reserve."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["token_budget"],
        max_context_tokens=1000,
        token_budget_reserve=500,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "x" * 2000},
        {"role": "assistant", "content": "y" * 2000},
        {"role": "user", "content": "z"},
    ]
    result = engine.optimize(msgs, max_tokens=1000)
    assert result.final_tokens <= 500


def test_budget_zero_disabled() -> None:
    """budget=0 → no trimming."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["token_budget"],
        max_context_tokens=0,
    )
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": "x" * 10000}]
    result = engine.optimize(msgs, max_tokens=0)
    assert len(result.messages) == 1


def test_budget_extreme_over() -> None:
    """Way over budget → trims to minimum (preserving last user)."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["token_budget"],
        max_context_tokens=20,
        token_budget_reserve=0,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "x" * 500},
        {"role": "assistant", "content": "y" * 500},
        {"role": "user", "content": "z" * 500},
    ]
    result = engine.optimize(msgs, max_tokens=50)
    assert len(result.messages) == 1
    assert "token_budget" in result.strategies_applied
    assert result.final_tokens < result.original_tokens


def test_budget_single_message() -> None:
    """One message over budget → kept (can't remove)."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["token_budget"],
        max_context_tokens=10,
        token_budget_reserve=0,
    )
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": "x" * 1000}]
    result = engine.optimize(msgs, max_tokens=50)
    assert len(result.messages) == 1


# --- Token Estimation (5 tests) ---


def test_estimate_tokens_string() -> None:
    """'hello world' → ~3 tokens."""
    est = ContextEngine._estimate_tokens([{"role": "user", "content": "hello world"}])
    assert 2 <= est <= 6


def test_estimate_tokens_content_blocks() -> None:
    """List content → correct sum."""
    msgs = [
        {
            "role": "user",
            "content": [{"type": "text", "text": "a" * 40}, {"type": "text", "text": "b" * 40}],
        }
    ]
    est = ContextEngine._estimate_tokens(msgs)
    assert est >= 20


def test_estimate_tokens_empty() -> None:
    """Empty messages → 0."""
    assert ContextEngine._estimate_tokens([]) == 0


def test_estimate_tokens_overhead() -> None:
    """Includes per-message overhead."""
    est_one = ContextEngine._estimate_tokens([{"role": "user", "content": ""}])
    est_three = ContextEngine._estimate_tokens(
        [{"role": "user", "content": ""}, {"role": "assistant", "content": ""}, {"role": "user", "content": ""}]
    )
    assert est_three > est_one


def test_hash_message_deterministic() -> None:
    """Same message → same hash."""
    msg = {"role": "user", "content": "hello"}
    h1 = ContextEngine._hash_message(msg)
    h2 = ContextEngine._hash_message(msg)
    assert h1 == h2
    assert len(h1) == 64


# --- Integration (7 tests) ---


def test_optimize_multiple_strategies() -> None:
    """Dedup + trim applied together."""
    cfg = ContextConfig(
        enabled=True,
        strategies=["dedup", "trim_tool_results"],
        max_tool_result_tokens=100,
    )
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "hi"},
        {"role": "user", "content": "hi"},
        {"role": "tool", "content": "x" * 5000},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) <= 3
    assert len(result.strategies_applied) >= 1


def test_optimize_disabled() -> None:
    """Engine disabled → passthrough."""
    cfg = ContextConfig(enabled=False, strategies=["dedup"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "user", "content": "a"},
        {"role": "user", "content": "a"},
    ]
    result = engine.optimize(msgs)
    assert len(result.messages) == 2
    assert result.strategies_applied == []


def test_optimize_empty_messages() -> None:
    """Empty list → no crash."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"])
    engine = ContextEngine(cfg)
    result = engine.optimize([])
    assert result.messages == []
    assert result.original_count == 0
    assert result.final_count == 0


def test_result_metrics_correct() -> None:
    """tokens_saved, counts correct."""
    cfg = ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=50)
    engine = ContextEngine(cfg)
    long_content = "x" * 5000
    msgs = [{"role": "tool", "content": long_content}]
    result = engine.optimize(msgs)
    assert result.original_count == 1
    assert result.final_count == 1
    assert result.tokens_saved > 0
    assert result.original_tokens > result.final_tokens


def test_get_stats_accumulation() -> None:
    """Stats increment across calls."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"], track_savings=True)
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": "x"}, {"role": "user", "content": "x"}]
    engine.optimize(msgs)
    engine.optimize(msgs)
    stats = engine.get_stats()
    assert stats["total_optimizations"] >= 2
    assert stats.get("strategy_hits", {}).get("dedup", 0) >= 2


def test_thread_safe_optimize() -> None:
    """10 concurrent threads."""
    cfg = ContextConfig(enabled=True, strategies=["dedup"], track_savings=True)
    engine = ContextEngine(cfg)
    msgs = [{"role": "user", "content": "a"}, {"role": "user", "content": "a"}]
    results: list[ContextResult] = []

    def run() -> None:
        r = engine.optimize(msgs)
        results.append(r)

    threads = [threading.Thread(target=run) for _ in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert len(results) == 10
    for r in results:
        assert len(r.messages) == 1


def test_context_result_strategies_applied() -> None:
    """Tracks which strategies fired."""
    cfg = ContextConfig(enabled=True, strategies=["dedup", "trim_system_dups"])
    engine = ContextEngine(cfg)
    msgs = [
        {"role": "system", "content": "S"},
        {"role": "system", "content": "S"},
        {"role": "user", "content": "u"},
        {"role": "user", "content": "u"},
    ]
    result = engine.optimize(msgs)
    assert "dedup" in result.strategies_applied or "trim_system_dups" in result.strategies_applied


# --- Proxy Integration (5 tests) ---


def test_proxy_phase_context() -> None:
    """Messages modified before upstream."""
    from orchesis.config import load_policy
    from orchesis.proxy import LLMHTTPProxy
    from pathlib import Path

    base = Path(__file__).parent.parent
    policy_path = base / "examples" / "orchesis.yaml"
    if not policy_path.exists():
        policy_path = base / "orchesis.yaml"
    if not policy_path.exists():
        policy_path = base / "examples" / "policy.yaml"
    if not policy_path.exists():
        policy_path = base / "policy.yaml"
    if not policy_path.exists():
        pytest.skip("No policy yaml found")
    try:
        policy = load_policy(policy_path)
    except Exception:
        pytest.skip("Could not load policy")
    policy["context_engine"] = {
        "enabled": True,
        "strategies": ["dedup", "trim_tool_results"],
        "max_tool_result_tokens": 100,
    }
    proxy = LLMHTTPProxy(policy_path=None)
    proxy._policy = policy
    proxy._context_engine = ContextEngine(
        ContextConfig(
            enabled=True,
            strategies=["dedup", "trim_tool_results"],
            max_tool_result_tokens=100,
        )
    )
    from orchesis.proxy import _RequestContext
    from http.server import BaseHTTPRequestHandler

    class FakeHandler:
        pass

    ctx = _RequestContext(handler=FakeHandler(), body={"messages": [{"role": "user", "content": "x"}, {"role": "user", "content": "x"}]})
    ok = proxy._phase_context(ctx)
    assert ok is True
    assert len(ctx.body["messages"]) == 1


def test_proxy_context_headers() -> None:
    """X-Orchesis-Context-Tokens-Saved header set."""
    from orchesis.proxy import LLMHTTPProxy, _RequestContext
    from http.server import BaseHTTPRequestHandler

    class FakeHandler:
        pass

    proxy = LLMHTTPProxy(policy_path=None)
    proxy._context_engine = ContextEngine(
        ContextConfig(enabled=True, strategies=["trim_tool_results"], max_tool_result_tokens=50)
    )
    long_content = "x" * 5000
    ctx = _RequestContext(
        handler=FakeHandler(),
        body={"messages": [{"role": "tool", "content": long_content}], "model": "", "max_tokens": 0},
    )
    proxy._phase_context(ctx)
    assert ctx.context_tokens_saved > 0
    assert "trim_tool_results" in ctx.context_strategies


def test_proxy_context_disabled() -> None:
    """No context engine → passthrough."""
    from orchesis.proxy import LLMHTTPProxy, _RequestContext

    class FakeHandler:
        pass

    proxy = LLMHTTPProxy(policy_path=None)
    proxy._context_engine = None
    msgs = [{"role": "user", "content": "a"}, {"role": "user", "content": "a"}]
    ctx = _RequestContext(handler=FakeHandler(), body={"messages": msgs})
    ok = proxy._phase_context(ctx)
    assert ok is True
    assert ctx.body["messages"] == msgs


def test_proxy_context_stats() -> None:
    """Stats includes context_engine section."""
    from orchesis.proxy import LLMHTTPProxy

    proxy = LLMHTTPProxy(policy_path=None)
    proxy._context_engine = ContextEngine(ContextConfig(enabled=True, strategies=["dedup"]))
    stats = proxy.stats
    assert "context_engine" in stats
    assert "total_tokens_saved" in stats["context_engine"]


def test_config_normalization() -> None:
    """context_engine config validated."""
    from orchesis.config import load_policy
    from pathlib import Path
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        f.write(
            """
context_engine:
  enabled: true
  strategies:
    - dedup
    - trim_tool_results
  max_context_tokens: 0
  token_budget_reserve: 4096
  sliding_window_size: 0
  preserve_system: true
  max_tool_result_tokens: 2000
  dedup_window: 50
  track_savings: true
"""
        )
        path = f.name
    try:
        policy = load_policy(path)
        assert "context_engine" in policy
        ce = policy["context_engine"]
        assert isinstance(ce, dict)
        assert ce.get("enabled") is True
        assert "dedup" in (ce.get("strategies") or [])
    finally:
        import os
        os.unlink(path)



"""Auto-generated regression tests from attack corpus.
DO NOT EDIT MANUALLY. Regenerate with: orchesis corpus --generate-tests
"""

import json
from pathlib import Path

import pytest

from orchesis.config import load_policy
from orchesis.engine import evaluate

CORPUS_DIR = Path(__file__).parent / "corpus"
POLICY_PATH = Path(__file__).parent.parent / "examples" / "production_policy.yaml"


def _load_corpus_entries():
    entries = []
    for file_path in sorted(CORPUS_DIR.glob("BYPASS-*.json")):
        with file_path.open(encoding="utf-8") as handle:
            entries.append(json.load(handle))
    return entries


@pytest.mark.parametrize("entry", _load_corpus_entries(), ids=lambda e: e["id"])
def test_corpus_regression(entry):
    """Verify that known bypass is now correctly handled."""
    policy = load_policy(POLICY_PATH)
    decision = evaluate(entry["request"], policy)
    assert not decision.allowed, (
        f"{entry['id']} ({entry['category']}): "
        f"expected DENY but got ALLOW. "
        f"Mutation: {entry['mutation']}"
    )



from __future__ import annotations

import hashlib
import json
import os
from pathlib import Path

from tests.cli_test_utils import CliRunner

from orchesis.cli import main
from orchesis.integrity import IntegrityMonitor, build_integrity_alert_callback
from orchesis.integrations import SlackNotifier


def test_init_creates_baseline_file_structure(tmp_path: Path) -> None:
    policy = tmp_path / "policy.yaml"
    policy.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    report = monitor.init([str(policy)])
    assert report.files_count == 1
    payload = json.loads((tmp_path / ".orchesis/integrity.json").read_text(encoding="utf-8"))
    assert payload["version"] == "1.0"
    assert isinstance(payload["files"], dict)


def test_init_directory_recursively_finds_configs(tmp_path: Path) -> None:
    cfg_dir = tmp_path / "cfg"
    cfg_dir.mkdir()
    (cfg_dir / "a.json").write_text("{}", encoding="utf-8")
    (cfg_dir / "b.yaml").write_text("x: 1\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    report = monitor.init([str(cfg_dir)])
    assert report.files_count == 2


def test_init_ignores_non_config_files(tmp_path: Path) -> None:
    cfg_dir = tmp_path / "cfg"
    cfg_dir.mkdir()
    (cfg_dir / "cache.pyc").write_bytes(b"x")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    report = monitor.init([str(cfg_dir)])
    assert report.files_count == 0


def test_auto_discover_finds_existing_paths(tmp_path: Path, monkeypatch) -> None:
    home = tmp_path / "home"
    (home / ".cursor").mkdir(parents=True)
    (home / ".cursor/mcp.json").write_text("{}", encoding="utf-8")
    monkeypatch.setattr("pathlib.Path.home", lambda: home)
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    found = monitor.auto_discover()
    assert any(item.endswith("mcp.json") for item in found)


def test_check_detects_modified_file(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    file_path.write_text("rules:\n  - name: budget_limit\n", encoding="utf-8")
    report = monitor.check()
    assert len(report.modified) == 1


def test_check_detects_removed_file(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    file_path.unlink()
    report = monitor.check()
    assert len(report.removed) == 1


def test_check_detects_added_file_in_monitored_dir(tmp_path: Path) -> None:
    cfg = tmp_path / "cfg"
    cfg.mkdir()
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(cfg)])
    (cfg / "new.json").write_text("{}", encoding="utf-8")
    report = monitor.check()
    assert len(report.added) == 1


def test_check_detects_permission_change(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    if os.name == "posix":
        os.chmod(file_path, 0o600)
    else:
        # emulate by mutating baseline for windows-friendly test
        payload = json.loads((tmp_path / ".orchesis/integrity.json").read_text(encoding="utf-8"))
        key = str(file_path.resolve())
        payload["files"][key]["permissions"] = "0000"
        (tmp_path / ".orchesis/integrity.json").write_text(json.dumps(payload), encoding="utf-8")
    report = monitor.check()
    assert len(report.permission_changed) == 1


def test_check_clean_report_when_unchanged(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    report = monitor.check()
    assert report.has_changes is False
    assert report.unchanged >= 1


def test_update_single_file_refreshes_baseline(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    file_path.write_text("rules:\n  - name: x\n", encoding="utf-8")
    _ = monitor.update([str(file_path)])
    report = monitor.check()
    assert report.has_changes is False


def test_update_all_files_refreshes_baseline(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    file_path.write_text("rules:\n  - name: x\n", encoding="utf-8")
    _ = monitor.update()
    report = monitor.check()
    assert report.has_changes is False


def test_cli_check_strict_exit_code(tmp_path: Path) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("policy.yaml").write_text("rules: []\n", encoding="utf-8")
        _ = runner.invoke(main, ["integrity", "init", "--paths", "policy.yaml"])
        Path("policy.yaml").write_text("rules:\n  - name: x\n", encoding="utf-8")
        result = runner.invoke(main, ["integrity", "check", "--strict"])
        assert result.exit_code == 1


def test_baseline_file_is_valid_json(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    loaded = json.loads((tmp_path / ".orchesis/integrity.json").read_text(encoding="utf-8"))
    assert isinstance(loaded, dict)


def test_sha256_computation_known_value(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    content = "abc"
    file_path.write_text(content, encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    payload = json.loads((tmp_path / ".orchesis/integrity.json").read_text(encoding="utf-8"))
    expected = hashlib.sha256(content.encode("utf-8")).hexdigest()
    actual = payload["files"][str(file_path.resolve())]["sha256"]
    assert actual == expected


def test_watch_calls_callback_on_change(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    file_path.write_text("rules:\n  - name: x\n", encoding="utf-8")
    seen = []
    monitor.watch(1, lambda report: seen.append(report), max_iterations=1)
    assert len(seen) == 1


def test_watch_no_callback_when_clean(tmp_path: Path) -> None:
    file_path = tmp_path / "policy.yaml"
    file_path.write_text("rules: []\n", encoding="utf-8")
    monitor = IntegrityMonitor(baseline_path=str(tmp_path / ".orchesis/integrity.json"))
    _ = monitor.init([str(file_path)])
    seen = []
    monitor.watch(1, lambda report: seen.append(report), max_iterations=1)
    assert seen == []


def test_alert_callback_sends_messages(monkeypatch) -> None:
    sent: list[str] = []
    monkeypatch.setattr(SlackNotifier, "send", lambda self, message: sent.append(message))
    policy = {"alerts": {"slack": {"webhook_url": "https://hooks.slack.com/services/T/B/X"}}}
    callback = build_integrity_alert_callback(policy)
    from orchesis.integrity import FileChange, IntegrityReport

    report = IntegrityReport(
        modified=[FileChange(path="policy.yaml", change_type="hash", old_value="a", new_value="b")],
        added=[],
        removed=[],
        permission_changed=[],
        unchanged=0,
    )
    callback(report)
    assert sent


def test_cli_integrity_status_and_update(tmp_path: Path) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("policy.yaml").write_text("rules: []\n", encoding="utf-8")
        _ = runner.invoke(main, ["integrity", "init", "--paths", "policy.yaml"])
        status = runner.invoke(main, ["integrity", "status"])
        assert status.exit_code == 0
        assert "Files tracked" in status.output
        update = runner.invoke(main, ["integrity", "update", "--path", "policy.yaml"])
        assert update.exit_code == 0


def test_integrity_init_paths_space_separated_values() -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("a").mkdir()
        Path("b").mkdir()
        Path("policy.yaml").write_text("rules: []\n", encoding="utf-8")
        Path("a/cfg.json").write_text("{}", encoding="utf-8")
        Path("b/cfg.yaml").write_text("x: 1\n", encoding="utf-8")
        result = runner.invoke(main, ["integrity", "init", "--paths", "a", "b", "policy.yaml"])
        assert result.exit_code == 0
        assert "Baseline created:" in result.output


def test_integrity_init_paths_single_value_backward_compat() -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("policy.yaml").write_text("rules: []\n", encoding="utf-8")
        result = runner.invoke(main, ["integrity", "init", "--paths", "policy.yaml"])
        assert result.exit_code == 0
        assert "Baseline created:" in result.output


def test_integrity_init_no_paths_no_autodiscover_shows_hint() -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["integrity", "init"])
    assert result.exit_code != 0
    assert "--paths" in result.output or "--auto-discover" in result.output



from __future__ import annotations

import asyncio
import json

import pytest

from orchesis.models import Decision
from orchesis.proxy import OrchesisProxy, ProxyConfig


class _Engine:
    def evaluate(self, payload: dict):
        _ = payload
        return Decision(allowed=True, reasons=[], rules_checked=[])


def _policy(block_request: bool = True, redact: bool = True) -> dict:
    return {
        "proxy": {
            "scan_requests": True,
            "scan_responses": True,
            "secret_scanning": {
                "enabled": True,
                "severity_threshold": "high",
                "block_on_critical": block_request,
            },
            "pii_scanning": {
                "enabled": True,
                "severity_threshold": "medium",
                "block_on_critical": False,
            },
            "response_redaction": {
                "enabled": redact,
                "redact_secrets": True,
                "redact_pii": False,
            },
        }
    }


def test_proxy_scans_request_for_secrets() -> None:
    proxy = OrchesisProxy(_Engine(), ProxyConfig(upstream_url="http://localhost:3000"), policy=_policy())
    findings = proxy._scan_request("write_file", {"token": "sk-abcdefghijklmnopqrstuvwxyz123"})
    assert findings


@pytest.mark.asyncio
async def test_proxy_blocks_critical_secret_in_request() -> None:
    proxy = OrchesisProxy(_Engine(), ProxyConfig(upstream_url="http://localhost:3000"), policy=_policy(block_request=True))

    async def _forward(method, path, headers, body):
        _ = (method, path, headers, body)
        raise AssertionError("must not forward blocked request")

    proxy._forward_request = _forward  # type: ignore[method-assign]
    server = await asyncio.start_server(proxy.handle_request, "127.0.0.1", 0)
    port = server.sockets[0].getsockname()[1]
    body = json.dumps(
        {"tool_name": "write_file", "params": {"token": "sk-abcdefghijklmnopqrstuvwxyz123"}}
    ).encode("utf-8")
    reader, writer = await asyncio.open_connection("127.0.0.1", port)
    writer.write(
        (
            "POST /invoke HTTP/1.1\r\n"
            "Host: test\r\n"
            "Content-Type: application/json\r\n"
            f"Content-Length: {len(body)}\r\n\r\n"
        ).encode("utf-8")
        + body
    )
    await writer.drain()
    raw = await reader.read()
    writer.close()
    await writer.wait_closed()
    server.close()
    await server.wait_closed()
    assert b"HTTP/1.1 403 Forbidden" in raw
    assert b"credential_leak_in_request" in raw


def test_proxy_scans_response_for_secrets() -> None:
    proxy = OrchesisProxy(_Engine(), ProxyConfig(upstream_url="http://localhost:3000"), policy=_policy())
    findings = proxy._scan_response_findings("read_file", b"api key sk-abcdefghijklmnopqrstuvwxyz123")
    assert findings["secrets"]


def test_proxy_redacts_secrets_in_response() -> None:
    proxy = OrchesisProxy(_Engine(), ProxyConfig(upstream_url="http://localhost:3000"), policy=_policy(redact=True))
    body = b'{"result":"API key is sk-abcdefghijklmnopqrstuvwxyz123"}'
    redacted = proxy._redact_response_body(body)
    assert b"[REDACTED-openai_key]" in redacted


def test_proxy_pii_scan_response() -> None:
    proxy = OrchesisProxy(_Engine(), ProxyConfig(upstream_url="http://localhost:3000"), policy=_policy())
    findings = proxy._scan_response_findings("read_file", b"Email: john.doe@example.com")
    assert findings["pii"]


def test_proxy_stats_tracks_secrets() -> None:
    proxy = OrchesisProxy(_Engine(), ProxyConfig(upstream_url="http://localhost:3000"), policy=_policy())
    proxy.stats.record_detection(secrets_detected=2, secrets_blocked=1, pii_detected=3)
    payload = proxy.stats.to_dict()
    assert payload["secrets_detected"] == 2
    assert payload["secrets_blocked"] == 1
    assert payload["pii_detected"] == 3


def test_proxy_scan_config_from_policy() -> None:
    proxy = OrchesisProxy(
        _Engine(),
        ProxyConfig(upstream_url="http://localhost:3000"),
        policy={
            "proxy": {
                "scan_requests": False,
                "scan_responses": False,
                "secret_scanning": {"enabled": False},
                "pii_scanning": {"enabled": False},
            }
        },
    )
    assert proxy._scan_requests is False
    assert proxy._scan_responses is False
    assert proxy._secret_enabled is False
    assert proxy._pii_enabled is False



from __future__ import annotations

import asyncio
import json
from typing import Any

from orchesis.integrations.alert_manager import AlertManager
from orchesis.integrations.base import AlertEvent, BaseIntegration
from orchesis.integrations.discord import DiscordIntegration, format_blocked_embed
from orchesis.integrations.github import GitHubIntegration
from orchesis.integrations.slack import SlackIntegration, format_blocked_message, format_budget_message
from orchesis.integrations.telegram import TelegramBot, TelegramIntegration
from orchesis.integrations.webhook import WebhookIntegration
from orchesis.integrations import webhook as webhook_module


def _event(
    *,
    action: str = "blocked",
    severity: str = "high",
    agent_id: str = "openclaw_01",
    rule_id: str = "ORCH-PI-001",
    pattern: str = "ignore previous instructions",
    description: str = "prompt injection detected",
    remediation: str = "sanitize user prompt",
    metadata: dict[str, Any] | None = None,
) -> AlertEvent:
    return AlertEvent(
        action=action,
        severity=severity,
        agent_id=agent_id,
        rule_id=rule_id,
        pattern=pattern,
        description=description,
        remediation=remediation,
        timestamp="2026-03-16T09:14:32+00:00",
        metadata=metadata or {},
    )


def test_alert_manager_dispatch_to_telegram(monkeypatch) -> None:  # noqa: ANN001
    sent: list[str] = []
    monkeypatch.setattr(TelegramIntegration, "send", lambda self, event: sent.append(event.action) or True)
    manager = AlertManager({"alerts": {"telegram": {"enabled": True, "token": "t", "chat_id": "c", "on": ["blocked"]}}})
    manager.dispatch(_event(action="blocked"))
    assert sent == ["blocked"]


def test_alert_manager_dispatch_to_slack(monkeypatch) -> None:  # noqa: ANN001
    sent: list[str] = []
    monkeypatch.setattr(SlackIntegration, "send", lambda self, event: sent.append(event.action) or True)
    manager = AlertManager(
        {"alerts": {"slack": {"enabled": True, "webhook_url": "https://hooks.slack.com/x", "on": ["blocked"]}}}
    )
    manager.dispatch(_event(action="blocked"))
    assert sent == ["blocked"]


def test_alert_manager_dispatch_to_discord(monkeypatch) -> None:  # noqa: ANN001
    sent: list[str] = []
    monkeypatch.setattr(DiscordIntegration, "send", lambda self, event: sent.append(event.action) or True)
    manager = AlertManager(
        {"alerts": {"discord": {"enabled": True, "webhook_url": "https://discord.com/api/webhooks/x", "on": ["blocked"]}}}
    )
    manager.dispatch(_event(action="blocked"))
    assert sent == ["blocked"]


def test_alert_manager_skip_disabled_integration(monkeypatch) -> None:  # noqa: ANN001
    sent: list[str] = []
    monkeypatch.setattr(SlackIntegration, "send", lambda self, event: sent.append(event.action) or True)
    manager = AlertManager(
        {"alerts": {"slack": {"enabled": False, "webhook_url": "https://hooks.slack.com/x", "on": ["blocked"]}}}
    )
    manager.dispatch(_event(action="blocked"))
    assert sent == []


def test_alert_manager_filter_by_severity(monkeypatch) -> None:  # noqa: ANN001
    sent: list[str] = []
    monkeypatch.setattr(SlackIntegration, "send", lambda self, event: sent.append(event.severity) or True)
    manager = AlertManager(
        {
            "alerts": {
                "slack": {
                    "enabled": True,
                    "webhook_url": "https://hooks.slack.com/x",
                    "on": ["blocked"],
                    "min_severity": "high",
                }
            }
        }
    )
    manager.dispatch(_event(action="blocked", severity="medium"))
    manager.dispatch(_event(action="blocked", severity="critical"))
    assert sent == ["critical"]


def test_alert_manager_filter_by_event_type(monkeypatch) -> None:  # noqa: ANN001
    sent: list[str] = []
    monkeypatch.setattr(SlackIntegration, "send", lambda self, event: sent.append(event.action) or True)
    manager = AlertManager(
        {"alerts": {"slack": {"enabled": True, "webhook_url": "https://hooks.slack.com/x", "on": ["budget_exceeded"]}}}
    )
    manager.dispatch(_event(action="blocked"))
    manager.dispatch(_event(action="budget_exceeded"))
    assert sent == ["budget_exceeded"]


def test_alert_manager_never_crashes_proxy(monkeypatch) -> None:  # noqa: ANN001
    def _boom(self, event):  # noqa: ANN001
        raise RuntimeError("network down")

    monkeypatch.setattr(SlackIntegration, "send", _boom)
    manager = AlertManager(
        {"alerts": {"slack": {"enabled": True, "webhook_url": "https://hooks.slack.com/x", "on": ["blocked"]}}}
    )
    manager.dispatch(_event(action="blocked"))


def test_telegram_format_blocked() -> None:
    integration = TelegramIntegration({"enabled": True, "token": "t", "chat_id": "c"})
    text = integration.format_event(_event(action="blocked", severity="high"))
    assert "BLOCKED" in text
    assert "Agent: openclaw_01" in text
    assert "Rule: ORCH-PI-001" in text


def test_telegram_format_budget() -> None:
    integration = TelegramIntegration({"enabled": True, "token": "t", "chat_id": "c"})
    text = integration.format_event(
        _event(
            action="budget_exceeded",
            metadata={"spent_usd": 4.23, "limit_usd": 5.0, "rate_per_hour": 0.18, "projected_24h": 5.0, "eta": "27 min"},
        )
    )
    assert "BUDGET WARNING" in text
    assert "$4.23 / $5.00" in text


def test_telegram_command_status(monkeypatch) -> None:  # noqa: ANN001
    class _Proxy:
        def get_status_snapshot(self):
            return {"uptime_seconds": 3600, "requests": 1847, "blocked": 3, "saved_usd": 2.41}

    bot = TelegramBot("token", "chat", _Proxy())
    sent: list[str] = []
    monkeypatch.setattr(bot, "send", lambda text: sent.append(text) or asyncio.sleep(0))
    asyncio.run(bot._handle_command({"update_id": 1, "message": {"text": "/status"}}))
    assert sent and "ALL CLEAR" in sent[0]


def test_telegram_command_block_pattern(monkeypatch) -> None:  # noqa: ANN001
    class _Proxy:
        def __init__(self):
            self.value = ""

        def add_block_pattern(self, pattern: str):
            self.value = pattern

    proxy = _Proxy()
    bot = TelegramBot("token", "chat", proxy)
    sent: list[str] = []
    monkeypatch.setattr(bot, "send", lambda text: sent.append(text) or asyncio.sleep(0))
    asyncio.run(bot._handle_command({"update_id": 2, "message": {"text": "/block rm -rf"}}))
    assert proxy.value == "rm -rf"
    assert sent and "Blocked: rm -rf" in sent[0]


def test_telegram_command_pause_resume(monkeypatch) -> None:  # noqa: ANN001
    class _Proxy:
        def __init__(self):
            self.monitor = False

        def set_monitoring_only(self, value: bool):
            self.monitor = bool(value)

    proxy = _Proxy()
    bot = TelegramBot("token", "chat", proxy)
    sent: list[str] = []
    monkeypatch.setattr(bot, "send", lambda text: sent.append(text) or asyncio.sleep(0))
    asyncio.run(bot._handle_command({"update_id": 3, "message": {"text": "/pause"}}))
    assert proxy.monitor is True
    asyncio.run(bot._handle_command({"update_id": 4, "message": {"text": "/resume"}}))
    assert proxy.monitor is False
    assert any("paused" in item.lower() for item in sent)
    assert any("resumed" in item.lower() for item in sent)


def test_slack_format_blocked_message() -> None:
    payload = format_blocked_message(_event(action="BLOCKED", severity="high"))
    assert payload["blocks"][0]["text"]["text"].startswith("🚨 HIGH")
    fields = payload["blocks"][1]["fields"]
    assert any("Agent" in field["text"] for field in fields)


def test_slack_format_budget_message() -> None:
    payload = format_budget_message(
        _event(action="BUDGET_EXCEEDED", metadata={"spent": 2.0, "limit": 5.0, "projection_24h": 10.0})
    )
    text = payload["blocks"][0]["text"]["text"]
    assert "BUDGET_EXCEEDED" in text


def test_discord_format_embed_colors() -> None:
    high_payload = format_blocked_embed(_event(severity="high", action="BLOCKED"))
    critical_payload = format_blocked_embed(_event(severity="critical", action="BLOCKED"))
    assert high_payload["embeds"][0]["color"] == 0xFF6600
    assert critical_payload["embeds"][0]["color"] == 0xFF0000


def test_github_create_issue_format(monkeypatch) -> None:  # noqa: ANN001
    captured: dict[str, Any] = {}

    def _fake_api_post(self, path: str, payload: dict[str, Any]):  # noqa: ANN001
        captured["path"] = path
        captured["payload"] = payload
        return {"ok": True}

    monkeypatch.setattr(GitHubIntegration, "_api_post", _fake_api_post)
    integration = GitHubIntegration({"enabled": True, "token": "ghp_x", "repo": "owner/repo"})
    integration.create_security_issue(_event(severity="critical"))
    assert captured["path"] == "/issues"
    assert "Security Alert from Orchesis" in captured["payload"]["body"]
    assert "severity:critical" in captured["payload"]["labels"]


def test_webhook_hmac_signature() -> None:
    integration = WebhookIntegration({"enabled": True, "url": "https://example.com/hook", "secret": "s3cr3t"})
    payload = integration.build_payload(_event())
    body = json.dumps(payload, ensure_ascii=False, sort_keys=True).encode("utf-8")
    signature = integration._signature(body)
    assert signature.startswith("sha256=")


def test_webhook_payload_format() -> None:
    integration = WebhookIntegration({"enabled": True, "url": "https://example.com/hook"})
    payload = integration.build_payload(_event(action="anomaly", severity="medium"))
    assert payload["source"] == "orchesis"
    assert payload["action"] == "anomaly"
    assert payload["severity"] == "medium"


def test_base_should_alert_severity_filter() -> None:
    base = BaseIntegration({"enabled": True, "on": ["blocked"], "min_severity": "high"})
    assert base.should_alert(_event(action="blocked", severity="medium"), base.config) is False
    assert base.should_alert(_event(action="blocked", severity="critical"), base.config) is True


def test_base_should_alert_event_filter() -> None:
    base = BaseIntegration({"enabled": True, "on": ["budget_exceeded"], "min_severity": "low"})
    assert base.should_alert(_event(action="blocked"), base.config) is False
    assert base.should_alert(_event(action="budget_exceeded"), base.config) is True


def test_webhook_send_uses_signature_header(monkeypatch) -> None:  # noqa: ANN001
    captured: dict[str, Any] = {}

    class _Resp:
        status = 200

        def __enter__(self):  # noqa: ANN001
            return self

        def __exit__(self, exc_type, exc, tb):  # noqa: ANN001
            return False

    def _fake_urlopen(req, timeout: float):  # noqa: ANN001
        captured["headers"] = {k.lower(): v for k, v in req.headers.items()}
        captured["timeout"] = timeout
        return _Resp()

    monkeypatch.setattr(webhook_module, "urlopen", _fake_urlopen)
    integration = WebhookIntegration(
        {"enabled": True, "url": "https://example.com/hook", "secret": "abc", "timeout_seconds": 3}
    )
    assert integration.send(_event()) is True
    assert "x-orchesis-signature" in captured["headers"]
    assert captured["timeout"] == 3




from __future__ import annotations

from orchesis.metrics import MetricsCollector
from orchesis.telemetry import DecisionEvent


def _event(
    *, agent: str, tool: str, decision: str, duration: int = 10, reasons: list[str] | None = None
):
    return DecisionEvent(
        event_id=f"{agent}-{tool}-{decision}-{duration}",
        timestamp="2026-01-01T00:00:00+00:00",
        agent_id=agent,
        tool=tool,
        params_hash="hash",
        cost=0.1,
        decision=decision,
        reasons=reasons or [],
        rules_checked=[],
        rules_triggered=[],
        evaluation_order=[],
        evaluation_duration_us=duration,
        policy_version="v1",
        state_snapshot={"tool_counts": {}},
    )


def test_metrics_count_decisions() -> None:
    metrics = MetricsCollector()
    for _ in range(5):
        metrics.emit(_event(agent="a", tool="read_file", decision="ALLOW"))
    for _ in range(3):
        metrics.emit(_event(agent="a", tool="write_file", decision="DENY"))
    snapshot = metrics.snapshot()
    counters = snapshot["counters"]
    assert counters["orchesis_decisions_total|decision=ALLOW"] == 5
    assert counters["orchesis_decisions_total|decision=DENY"] == 3


def test_metrics_by_agent() -> None:
    metrics = MetricsCollector()
    metrics.emit(_event(agent="agent_a", tool="read_file", decision="ALLOW"))
    metrics.emit(_event(agent="agent_b", tool="read_file", decision="DENY"))
    counters = metrics.snapshot()["counters"]
    assert counters["orchesis_decisions_by_agent|agent=agent_a|decision=ALLOW"] == 1
    assert counters["orchesis_decisions_by_agent|agent=agent_b|decision=DENY"] == 1


def test_metrics_prometheus_format() -> None:
    metrics = MetricsCollector()
    metrics.emit(_event(agent="a", tool="read_file", decision="ALLOW"))
    text = metrics.prometheus_text()
    assert "# HELP orchesis_decisions_total" in text
    assert 'orchesis_decisions_total{decision="ALLOW"}' in text
    assert "# TYPE orchesis_evaluation_duration_us histogram" in text


def test_metrics_evaluation_duration() -> None:
    metrics = MetricsCollector()
    metrics.emit(_event(agent="a", tool="read_file", decision="ALLOW", duration=50))
    metrics.emit(_event(agent="a", tool="read_file", decision="ALLOW", duration=150))
    hist = metrics.snapshot()["histograms"]["orchesis_evaluation_duration_us"]
    assert hist == [50.0, 150.0]


def test_metrics_reset() -> None:
    metrics = MetricsCollector()
    metrics.emit(_event(agent="a", tool="read_file", decision="ALLOW"))
    metrics.reset()
    snapshot = metrics.snapshot()
    assert snapshot["counters"] == {}
    assert snapshot["histograms"] == {}



"""A/B Testing Framework and Task Completion Tracking for AI agent optimization."""

from __future__ import annotations

import hashlib
import json
import math
import random
import time
import threading
from collections import OrderedDict
from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Any, Optional


class ExperimentStatus(Enum):
    """Lifecycle of an experiment."""

    DRAFT = "draft"
    RUNNING = "running"
    PAUSED = "paused"
    COMPLETED = "completed"
    CANCELLED = "cancelled"


class TaskOutcome(Enum):
    """How a task/session ended."""

    SUCCESS = "success"
    FAILURE = "failure"
    ABANDONED = "abandoned"
    TIMEOUT = "timeout"
    LOOP = "loop"
    ESCALATED = "escalated"
    UNKNOWN = "unknown"


class SplitStrategy(Enum):
    """How to assign requests to variants."""

    RANDOM = "random"
    STICKY_SESSION = "sticky_session"
    STICKY_AGENT = "sticky_agent"
    ROUND_ROBIN = "round_robin"


@dataclass
class Variant:
    """One arm of an A/B experiment."""

    name: str
    weight: float = 0.5
    model_override: str = ""
    config_overrides: dict = field(default_factory=dict)
    requests: int = 0
    successes: int = 0
    failures: int = 0
    total_cost_usd: float = 0.0
    total_latency_ms: float = 0.0
    total_tokens: int = 0
    total_tool_calls: int = 0
    avg_turns: float = 0.0
    _turns_sum: int = 0
    _tasks_completed: int = 0
    _latency_reservoir: list[float] = field(default_factory=list)
    _reservoir_max: int = 1000


@dataclass
class VariantAssignment:
    """Result of assigning a request to an experiment variant."""

    experiment_id: str
    variant_name: str
    model_override: str
    config_overrides: dict


@dataclass
class VariantStats:
    """Computed stats for one variant."""

    name: str
    sample_size: int
    success_rate: float
    avg_cost_usd: float
    avg_latency_ms: float
    avg_tokens: int
    avg_tool_calls: float
    avg_turns: float
    p95_latency_ms: float
    error_rate: float
    cost_per_success: float


@dataclass
class ExperimentResult:
    """Statistical comparison of experiment variants."""

    experiment_id: str
    is_significant: bool
    winner: str
    confidence: float
    variants: dict[str, VariantStats]
    recommendation: str
    cost_comparison: str

    def to_dict(self) -> dict:
        return {
            "experiment_id": self.experiment_id,
            "is_significant": self.is_significant,
            "winner": self.winner,
            "confidence": self.confidence,
            "variants": {k: asdict(v) for k, v in self.variants.items()},
            "recommendation": self.recommendation,
            "cost_comparison": self.cost_comparison,
        }


@dataclass
class Experiment:
    """A/B experiment definition."""

    experiment_id: str
    name: str
    description: str = ""
    status: ExperimentStatus = ExperimentStatus.DRAFT
    variants: list[Variant] = field(default_factory=list)
    split_strategy: SplitStrategy = SplitStrategy.STICKY_SESSION
    target_models: list[str] = field(default_factory=list)
    target_agents: list[str] = field(default_factory=list)
    target_tools: list[str] = field(default_factory=list)
    max_requests: int = 0
    max_duration_seconds: float = 0.0
    min_sample_size: int = 30
    created_at: float = 0.0
    started_at: float = 0.0
    ended_at: float = 0.0
    _round_robin_counter: int = 0

    def to_dict(self) -> dict:
        return {
            "experiment_id": self.experiment_id,
            "name": self.name,
            "description": self.description,
            "status": self.status.value,
            "variants": [
                {
                    "name": v.name,
                    "weight": v.weight,
                    "model_override": v.model_override,
                    "config_overrides": v.config_overrides,
                    "requests": v.requests,
                    "successes": v.successes,
                    "failures": v.failures,
                    "total_cost_usd": v.total_cost_usd,
                    "total_latency_ms": v.total_latency_ms,
                    "total_tokens": v.total_tokens,
                    "total_tool_calls": v.total_tool_calls,
                    "avg_turns": v.avg_turns,
                }
                for v in self.variants
            ],
            "split_strategy": self.split_strategy.value,
            "target_models": self.target_models,
            "target_agents": self.target_agents,
            "target_tools": self.target_tools,
            "max_requests": self.max_requests,
            "max_duration_seconds": self.max_duration_seconds,
            "min_sample_size": self.min_sample_size,
            "created_at": self.created_at,
            "started_at": self.started_at,
            "ended_at": self.ended_at,
        }


@dataclass
class ExperimentConfig:
    """Configuration for experiments and task tracking."""

    max_experiments: int = 10
    default_min_sample_size: int = 30
    auto_stop_on_significance: bool = True
    significance_threshold: float = 0.95
    max_tracked_sessions: int = 5000
    idle_timeout_seconds: float = 300.0
    min_turns_for_success: int = 1
    consecutive_errors_threshold: int = 3


@dataclass
class TaskSession:
    """State of a tracked task/session."""

    session_id: str
    started_at: float
    last_activity: float
    turns: int = 0
    models_used: list[str] = field(default_factory=list)
    total_cost_usd: float = 0.0
    total_tokens: int = 0
    total_tool_calls: int = 0
    total_latency_ms: float = 0.0
    errors: int = 0
    consecutive_errors: int = 0
    loop_detected: bool = False
    was_escalated: bool = False
    stop_reasons: list[str] = field(default_factory=list)
    outcome: TaskOutcome = TaskOutcome.UNKNOWN
    experiment_id: str = ""
    variant_name: str = ""


class TaskCorrelations:
    """Accumulates correlation data between features and outcomes."""

    def __init__(self) -> None:
        self._by_model: dict[str, dict[str, Any]] = {}
        self._by_tool_count: dict[str, dict[str, Any]] = {}
        self._by_turn_count: dict[str, dict[str, Any]] = {}

    def _get_bucket(self, d: dict[str, dict], key: str) -> dict[str, Any]:
        if key not in d:
            d[key] = {"success": 0, "total": 0, "cost_sum": 0.0}
        return d[key]

    def record(self, session: TaskSession) -> None:
        success = session.outcome == TaskOutcome.SUCCESS
        model = session.models_used[-1] if session.models_used else "unknown"
        bucket_m = self._get_bucket(self._by_model, model)
        bucket_m["success"] += 1 if success else 0
        bucket_m["total"] += 1
        bucket_m["cost_sum"] += session.total_cost_usd

        tc = session.total_tool_calls
        tc_label = self._bucket_count(tc, [(2, "1-2"), (5, "3-5"), (999999, "5+")])
        bucket_tc = self._get_bucket(self._by_tool_count, tc_label)
        bucket_tc["success"] += 1 if success else 0
        bucket_tc["total"] += 1

        turns = session.turns
        turn_label = self._bucket_count(turns, [(3, "1-3"), (7, "4-7"), (999999, "8+")])
        bucket_t = self._get_bucket(self._by_turn_count, turn_label)
        bucket_t["success"] += 1 if success else 0
        bucket_t["total"] += 1

    @staticmethod
    def _bucket_count(n: int, thresholds: list[tuple[int, str]]) -> str:
        for thresh, label in thresholds:
            if n <= thresh:
                return label
        return thresholds[-1][1] if thresholds else "unknown"

    def compute(self) -> dict:
        by_model: dict[str, dict[str, Any]] = {}
        for model, data in self._by_model.items():
            total = data["total"]
            if total > 0:
                by_model[model] = {
                    "success_rate": data["success"] / total,
                    "avg_cost": data["cost_sum"] / total,
                    "sample": total,
                }

        by_tool_count: dict[str, dict[str, Any]] = {}
        for label, data in self._by_tool_count.items():
            total = data["total"]
            if total > 0:
                by_tool_count[label] = {"success_rate": data["success"] / total, "sample": total}

        by_turn_count: dict[str, dict[str, Any]] = {}
        for label, data in self._by_turn_count.items():
            total = data["total"]
            if total > 0:
                by_turn_count[label] = {"success_rate": data["success"] / total, "sample": total}

        insights: list[str] = []
        if len(by_tool_count) >= 2:
            sorted_tc = sorted(by_tool_count.items(), key=lambda x: x[1]["success_rate"])
            worst = sorted_tc[0]
            best = sorted_tc[-1]
            if worst[1]["sample"] >= 10 and best[1]["sample"] >= 10:
                diff = best[1]["success_rate"] - worst[1]["success_rate"]
                if diff >= 0.15:
                    insights.append(
                        f"Tasks with {worst[0]} tool calls have {worst[1]['success_rate']:.0%} "
                        f"success rate vs {best[1]['success_rate']:.0%} for {best[0]}"
                    )
        if len(by_model) >= 2:
            sorted_m = sorted(by_model.items(), key=lambda x: x[1]["success_rate"])
            worst_m = sorted_m[0]
            best_m = sorted_m[-1]
            if worst_m[1]["sample"] >= 20 and best_m[1]["sample"] >= 20:
                diff = best_m[1]["success_rate"] - worst_m[1]["success_rate"]
                if diff >= 0.1:
                    insights.append(
                        f"{worst_m[0]} success rate is {worst_m[1]['success_rate']:.0%} vs "
                        f"{best_m[1]['success_rate']:.0%} for {best_m[0]}"
                    )
        if len(by_turn_count) >= 2:
            sorted_turns = sorted(by_turn_count.items(), key=lambda x: x[1]["success_rate"])
            worst_t = sorted_turns[0]
            best_t = sorted_turns[-1]
            if worst_t[1]["sample"] >= 10 and best_t[1]["sample"] >= 10:
                diff = best_t[1]["success_rate"] - worst_t[1]["success_rate"]
                if diff >= 0.15:
                    insights.append(
                        f"Sessions with {worst_t[0]} turns have {worst_t[1]['success_rate']:.0%} "
                        f"success rate vs {best_t[1]['success_rate']:.0%} for {best_t[0]}"
                    )

        return {
            "by_model": by_model,
            "by_tool_count": by_tool_count,
            "by_turn_count": by_turn_count,
            "insights": insights[:10],
        }


class TaskTracker:
    """Proxy-level task completion tracking."""

    def __init__(self, config: Optional[ExperimentConfig] = None) -> None:
        self._config = config or ExperimentConfig()
        self._lock = threading.Lock()
        self._sessions: OrderedDict[str, TaskSession] = OrderedDict()
        self._outcome_stats: dict[str, int] = {}
        self._correlations = TaskCorrelations()
        self._max_sessions = self._config.max_tracked_sessions

    def record_turn(
        self,
        session_id: str,
        model: str,
        tokens_in: int,
        tokens_out: int,
        cost_usd: float,
        latency_ms: float,
        tool_calls: int,
        stop_reason: str,
        is_error: bool,
        was_escalated: bool = False,
        was_loop_detected: bool = False,
        experiment_id: str = "",
        variant_name: str = "",
    ) -> None:
        now = time.time()
        with self._lock:
            if session_id not in self._sessions:
                if len(self._sessions) >= self._max_sessions:
                    self._sessions.popitem(last=False)
                self._sessions[session_id] = TaskSession(
                    session_id=session_id,
                    started_at=now,
                    last_activity=now,
                )
            s = self._sessions[session_id]
            s.last_activity = now
            s.turns += 1
            if model and model not in s.models_used:
                s.models_used.append(model)
            s.total_cost_usd += cost_usd
            s.total_tokens += tokens_in + tokens_out
            s.total_tool_calls += tool_calls
            s.total_latency_ms += latency_ms
            if stop_reason:
                s.stop_reasons.append(stop_reason)
            if is_error:
                s.errors += 1
                s.consecutive_errors += 1
            else:
                s.consecutive_errors = 0
            if was_loop_detected:
                s.loop_detected = True
            if was_escalated:
                s.was_escalated = True
            if experiment_id:
                s.experiment_id = experiment_id
            if variant_name:
                s.variant_name = variant_name

    def finalize_session(self, session_id: str) -> TaskOutcome:
        with self._lock:
            s = self._sessions.get(session_id)
            if s is None:
                return TaskOutcome.UNKNOWN
            if s.loop_detected:
                outcome = TaskOutcome.LOOP
            elif s.consecutive_errors >= self._config.consecutive_errors_threshold:
                outcome = TaskOutcome.FAILURE
            elif s.was_escalated:
                outcome = TaskOutcome.ESCALATED
            elif s.turns >= self._config.min_turns_for_success:
                last_stop = s.stop_reasons[-1] if s.stop_reasons else ""
                if last_stop == "end_turn" and s.total_tool_calls > 0:
                    outcome = TaskOutcome.SUCCESS
                else:
                    outcome = TaskOutcome.UNKNOWN
            elif s.turns < self._config.min_turns_for_success and not s.stop_reasons:
                outcome = TaskOutcome.ABANDONED
            else:
                outcome = TaskOutcome.UNKNOWN
            s.outcome = outcome
            self._outcome_stats[outcome.value] = self._outcome_stats.get(outcome.value, 0) + 1
            self._correlations.record(s)
            return outcome

    def get_session_state(self, session_id: str) -> Optional[TaskSession]:
        with self._lock:
            return self._sessions.get(session_id)

    def get_outcome_distribution(self) -> dict[str, int]:
        with self._lock:
            return dict(self._outcome_stats)

    def get_correlations(self) -> dict:
        with self._lock:
            return self._correlations.compute()

    def get_stats(self) -> dict:
        with self._lock:
            total = sum(self._outcome_stats.values())
            successes = self._outcome_stats.get(TaskOutcome.SUCCESS.value, 0)
            return {
                "tracked_sessions": len(self._sessions),
                "outcomes": dict(self._outcome_stats),
                "overall_success_rate": successes / total if total > 0 else 0.0,
            }


class ExperimentManager:
    """Manages A/B experiments and task completion tracking."""

    def __init__(self, config: Optional[ExperimentConfig] = None) -> None:
        self._config = config or ExperimentConfig()
        self._lock = threading.RLock()
        self._experiments: dict[str, Experiment] = {}
        self._task_tracker = TaskTracker(config)
        self._rng = random.Random()

    def create_experiment(
        self,
        name: str,
        variants: list[dict],
        split_strategy: str = "sticky_session",
        **kwargs: Any,
    ) -> Experiment:
        with self._lock:
            if len(self._experiments) >= self._config.max_experiments:
                raise ValueError("max_experiments limit reached")
            exp_id = hashlib.sha256(f"{name}{time.time()}".encode()).hexdigest()[:16]
            parsed_variants: list[Variant] = []
            for v in variants:
                if isinstance(v, dict):
                    parsed_variants.append(
                        Variant(
                            name=str(v.get("name", "control")),
                            weight=float(v.get("weight", 0.5)),
                            model_override=str(v.get("model_override", "")),
                            config_overrides=dict(v.get("config_overrides", {})),
                        )
                    )
            total_weight = sum(v.weight for v in parsed_variants)
            if total_weight <= 0:
                total_weight = 1.0
            for v in parsed_variants:
                v.weight = v.weight / total_weight
            strategy = SplitStrategy.RANDOM
            for s in SplitStrategy:
                if s.value == split_strategy:
                    strategy = s
                    break
            exp = Experiment(
                experiment_id=exp_id,
                name=name,
                description=str(kwargs.get("description", "")),
                status=ExperimentStatus.DRAFT,
                variants=parsed_variants,
                split_strategy=strategy,
                target_models=list(kwargs.get("target_models", []) or []),
                target_agents=list(kwargs.get("target_agents", []) or []),
                target_tools=list(kwargs.get("target_tools", []) or []),
                max_requests=int(kwargs.get("max_requests", 0) or 0),
                max_duration_seconds=float(kwargs.get("max_duration_seconds", 0) or 0),
                min_sample_size=int(kwargs.get("min_sample_size", self._config.default_min_sample_size) or 30),
                created_at=time.time(),
            )
            self._experiments[exp_id] = exp
            return exp

    def start_experiment(self, experiment_id: str) -> bool:
        with self._lock:
            exp = self._experiments.get(experiment_id)
            if exp is None:
                return False
            if exp.status == ExperimentStatus.RUNNING:
                return False
            if exp.status not in (ExperimentStatus.DRAFT, ExperimentStatus.PAUSED):
                return False
            exp.status = ExperimentStatus.RUNNING
            exp.started_at = time.time()
            return True

    def stop_experiment(self, experiment_id: str) -> ExperimentResult:
        with self._lock:
            exp = self._experiments.get(experiment_id)
            if exp is None:
                raise ValueError("experiment not found")
            exp.status = ExperimentStatus.COMPLETED
            exp.ended_at = time.time()
        return self.get_results(experiment_id)

    def pause_experiment(self, experiment_id: str) -> bool:
        with self._lock:
            exp = self._experiments.get(experiment_id)
            if exp is None or exp.status != ExperimentStatus.RUNNING:
                return False
            exp.status = ExperimentStatus.PAUSED
            return True

    def resume_experiment(self, experiment_id: str) -> bool:
        return self.start_experiment(experiment_id)

    def delete_experiment(self, experiment_id: str) -> bool:
        with self._lock:
            if experiment_id in self._experiments:
                del self._experiments[experiment_id]
                return True
            return False

    def list_experiments(self) -> list[dict]:
        with self._lock:
            return [e.to_dict() for e in self._experiments.values()]

    def get_experiment(self, experiment_id: str) -> Optional[Experiment]:
        with self._lock:
            return self._experiments.get(experiment_id)

    def assign_variant(
        self,
        session_id: str,
        agent_id: str,
        model: str,
        tools: list[str],
    ) -> Optional[VariantAssignment]:
        with self._lock:
            active = [
                e
                for e in self._experiments.values()
                if e.status == ExperimentStatus.RUNNING
            ]
            if not active:
                return None
            exp = active[0]
            if exp.target_models and model not in exp.target_models:
                return None
            if exp.target_agents and agent_id not in exp.target_agents:
                return None
            if exp.target_tools:
                if not any(t in exp.target_tools for t in tools):
                    return None
            variant = self._weighted_choice(exp, session_id, agent_id)
            if variant is None:
                return None
            return VariantAssignment(
                experiment_id=exp.experiment_id,
                variant_name=variant.name,
                model_override=variant.model_override or "",
                config_overrides=variant.config_overrides or {},
            )

    def _weighted_choice(self, exp: Experiment, session_id: str, agent_id: str) -> Optional[Variant]:
        if not exp.variants:
            return None
        if exp.split_strategy == SplitStrategy.ROUND_ROBIN:
            exp._round_robin_counter += 1
            idx = (exp._round_robin_counter - 1) % len(exp.variants)
            return exp.variants[idx]
        seed = ""
        if exp.split_strategy == SplitStrategy.STICKY_SESSION:
            seed = session_id
        elif exp.split_strategy == SplitStrategy.STICKY_AGENT:
            seed = agent_id
        else:
            seed = hashlib.sha256(f"{time.time()}{session_id}{agent_id}".encode()).hexdigest()
        h = int(hashlib.sha256(seed.encode()).hexdigest(), 16) % 10000
        r = (h / 10000.0) if seed else self._rng.random()
        cumul = 0.0
        for v in exp.variants:
            cumul += v.weight
            if r < cumul:
                return v
        return exp.variants[-1]

    def record_request(
        self,
        experiment_id: str,
        variant_name: str,
        cost_usd: float,
        latency_ms: float,
        tokens: int,
        tool_calls: int,
        is_error: bool,
        turns: int = 1,
    ) -> None:
        """Record per-request metrics. Does NOT update success/failure counters —
        those come from record_task_outcome() when session finalizes."""
        with self._lock:
            exp = self._experiments.get(experiment_id)
            if exp is None:
                return
            for v in exp.variants:
                if v.name == variant_name:
                    v.requests += 1
                    v.total_cost_usd += cost_usd
                    v.total_latency_ms += latency_ms
                    v._latency_reservoir.append(latency_ms)
                    if len(v._latency_reservoir) > v._reservoir_max:
                        v._latency_reservoir.pop(0)
                    v.total_tokens += tokens
                    v.total_tool_calls += tool_calls
                    if is_error:
                        v.failures += 1
                    v._turns_sum += turns
                    v._tasks_completed += 1
                    if v._tasks_completed > 0:
                        v.avg_turns = v._turns_sum / v._tasks_completed
                    if self._check_auto_stop(exp):
                        exp.status = ExperimentStatus.COMPLETED
                        exp.ended_at = time.time()
                    break

    def record_task_outcome(
        self,
        experiment_id: str,
        variant_name: str,
        outcome: TaskOutcome,
    ) -> None:
        """Update variant success/failure from task completion outcome."""
        with self._lock:
            exp = self._experiments.get(experiment_id)
            if exp is None:
                return
            for v in exp.variants:
                if v.name == variant_name:
                    if outcome == TaskOutcome.SUCCESS:
                        v.successes += 1
                    elif outcome in (
                        TaskOutcome.FAILURE,
                        TaskOutcome.LOOP,
                        TaskOutcome.TIMEOUT,
                        TaskOutcome.ABANDONED,
                    ):
                        v.failures += 1
                    break

    def _check_auto_stop(self, exp: Experiment) -> bool:
        if exp.max_requests > 0:
            total = sum(v.requests for v in exp.variants)
            if total >= exp.max_requests:
                return True
        if exp.max_duration_seconds > 0 and exp.started_at > 0:
            if time.time() - exp.started_at >= exp.max_duration_seconds:
                return True
        if self._config.auto_stop_on_significance and len(exp.variants) >= 2:
            r = self.get_results(exp.experiment_id)
            if r.is_significant:
                return True
        return False

    @staticmethod
    def _z_test_proportions(
        successes1: int,
        n1: int,
        successes2: int,
        n2: int,
    ) -> tuple[float, float]:
        if n1 <= 0 or n2 <= 0:
            return 0.0, 1.0
        p1 = successes1 / n1
        p2 = successes2 / n2
        p_pooled = (successes1 + successes2) / (n1 + n2)
        if p_pooled <= 0 or p_pooled >= 1:
            return 0.0, 1.0
        denom = math.sqrt(p_pooled * (1 - p_pooled) * (1 / n1 + 1 / n2))
        if denom <= 0:
            return 0.0, 1.0
        z = (p1 - p2) / denom
        p_value = 2 * (1 - _normal_cdf(abs(z)))
        return z, p_value

    def get_results(self, experiment_id: str) -> ExperimentResult:
        with self._lock:
            exp = self._experiments.get(experiment_id)
            if exp is None:
                raise ValueError("experiment not found")
        variants_stats: dict[str, VariantStats] = {}
        for v in exp.variants:
            n = v.requests
            if n == 0:
                variants_stats[v.name] = VariantStats(
                    name=v.name,
                    sample_size=0,
                    success_rate=0.0,
                    avg_cost_usd=0.0,
                    avg_latency_ms=0.0,
                    avg_tokens=0,
                    avg_tool_calls=0.0,
                    avg_turns=0.0,
                    p95_latency_ms=0.0,
                    error_rate=0.0,
                    cost_per_success=0.0,
                )
                continue
            sr = (v.successes / n) if n > 0 else 0.0
            err_rate = (v.failures / n) if n > 0 else 0.0
            avg_cost = (v.total_cost_usd / n) if n > 0 else 0.0
            avg_lat = (v.total_latency_ms / n) if n > 0 else 0.0
            avg_tok = int(v.total_tokens / n) if n > 0 else 0
            avg_tc = (v.total_tool_calls / n) if n > 0 else 0.0
            avg_t = v.avg_turns
            cps = (v.total_cost_usd / v.successes) if v.successes > 0 else 0.0
            reservoir = list(v._latency_reservoir)
            if reservoir:
                reservoir.sort()
                p95_idx = int(len(reservoir) * 0.95)
                p95 = reservoir[min(p95_idx, len(reservoir) - 1)]
            else:
                p95 = 0.0
            variants_stats[v.name] = VariantStats(
                name=v.name,
                sample_size=n,
                success_rate=sr,
                avg_cost_usd=avg_cost,
                avg_latency_ms=avg_lat,
                avg_tokens=avg_tok,
                avg_tool_calls=avg_tc,
                avg_turns=avg_t,
                p95_latency_ms=p95,
                error_rate=err_rate,
                cost_per_success=cps,
            )

        is_sig = False
        winner = ""
        confidence = 0.0
        if len(exp.variants) >= 2:
            v0, v1 = exp.variants[0], exp.variants[1]
            n0, n1 = v0.requests, v1.requests
            if n0 >= exp.min_sample_size and n1 >= exp.min_sample_size:
                z, p = self._z_test_proportions(v0.successes, n0, v1.successes, n1)
                confidence = 1.0 - p
                if p < (1 - self._config.significance_threshold):
                    is_sig = True
                    winner = v0.name if v0.successes / max(1, n0) > v1.successes / max(1, n1) else v1.name

        best = max(exp.variants, key=lambda v: (v.successes / max(1, v.requests), -v.total_cost_usd))
        if is_sig:
            rec = f"Statistically significant: {winner} is recommended (confidence {confidence:.2%})."
        else:
            rec = f"Insufficient data for statistical significance. Current leader: {best.name} ({best.successes}/{max(1, best.requests)} success rate)."
        cost_comp = ""
        if len(exp.variants) >= 2:
            v0, v1 = exp.variants[0], exp.variants[1]
            diff = (v1.total_cost_usd - v0.total_cost_usd) / max(1, v0.requests + v1.requests) * 1000
            cost_comp = f"Variant {v1.name} costs ${abs(diff):.4f} {'more' if diff > 0 else 'less'} per 1000 requests vs {v0.name}"
        return ExperimentResult(
            experiment_id=experiment_id,
            is_significant=is_sig,
            winner=winner,
            confidence=confidence,
            variants=variants_stats,
            recommendation=rec,
            cost_comparison=cost_comp,
        )

    def get_live_stats(self, experiment_id: str) -> dict:
        with self._lock:
            exp = self._experiments.get(experiment_id)
            if exp is None:
                return {}
        result = self.get_results(experiment_id)
        return {
            "experiment_id": experiment_id,
            "status": exp.status.value,
            "variants": {
                k: asdict(v) for k, v in result.variants.items()
            },
            "is_significant": result.is_significant,
            "winner": result.winner,
        }


def _normal_cdf(z: float) -> float:
    """Standard normal CDF using math.erf."""
    return 0.5 * (1 + math.erf(z / math.sqrt(2)))



"""Minimal MCP demo server for Orchesis proxy e2e checks."""

from __future__ import annotations

import asyncio
from typing import Any

import mcp.server.stdio
import mcp.types as types
from mcp.server.lowlevel import NotificationOptions, Server
from mcp.server.models import InitializationOptions

server = Server("demo-mcp-server")


@server.list_tools()
async def list_tools() -> list[types.Tool]:
    return [
        types.Tool(
            name="read_file",
            description="Read file content (stubbed)",
            inputSchema={
                "type": "object",
                "properties": {"path": {"type": "string"}},
                "required": ["path"],
            },
        ),
        types.Tool(
            name="delete_file",
            description="Delete file (stubbed)",
            inputSchema={
                "type": "object",
                "properties": {"path": {"type": "string"}},
                "required": ["path"],
            },
        ),
        types.Tool(
            name="run_sql",
            description="Run SQL query (stubbed)",
            inputSchema={
                "type": "object",
                "properties": {"query": {"type": "string"}},
                "required": ["query"],
            },
        ),
    ]


@server.call_tool()
async def call_tool(name: str, arguments: dict[str, Any]) -> list[types.TextContent]:
    if name == "read_file":
        path = str(arguments.get("path", ""))
        return [types.TextContent(type="text", text=f"fake content from {path}")]

    if name == "delete_file":
        return [types.TextContent(type="text", text="deleted")]

    if name == "run_sql":
        return [types.TextContent(type="text", text="executed")]

    raise ValueError(f"Unknown tool: {name}")


async def run() -> None:
    async with mcp.server.stdio.stdio_server() as (read_stream, write_stream):
        await server.run(
            read_stream,
            write_stream,
            InitializationOptions(
                server_name="demo-mcp-server",
                server_version="0.1.0",
                capabilities=server.get_capabilities(
                    notification_options=NotificationOptions(),
                    experimental_capabilities={},
                ),
            ),
        )


def main() -> None:
    asyncio.run(run())


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path

import yaml


ROOT = Path(__file__).resolve().parents[1]


def test_dockerfile_exists() -> None:
    assert (ROOT / "Dockerfile").exists()


def test_docker_compose_valid_yaml() -> None:
    compose_path = ROOT / "docker-compose.yml"
    data = yaml.safe_load(compose_path.read_text(encoding="utf-8"))
    assert isinstance(data, dict)


def test_docker_compose_has_services() -> None:
    compose_path = ROOT / "docker-compose.yml"
    data = yaml.safe_load(compose_path.read_text(encoding="utf-8"))
    services = data.get("services", {})
    assert "orchesis" in services


def test_makefile_exists() -> None:
    assert (ROOT / "Makefile").exists()


def test_makefile_has_key_targets() -> None:
    makefile = (ROOT / "Makefile").read_text(encoding="utf-8")
    assert "test:" in makefile
    assert "lint:" in makefile
    assert "fuzz:" in makefile
    assert "invariants:" in makefile
    assert "docker:" in makefile
    assert "help:" in makefile


def test_env_example_exists() -> None:
    assert (ROOT / ".env.example").exists()


def test_deployment_docs_exist() -> None:
    assert (ROOT / "docs" / "DEPLOYMENT.md").exists()


def test_dockerfile_uses_nonroot_user() -> None:
    dockerfile = (ROOT / "Dockerfile").read_text(encoding="utf-8")
    assert "FROM python:3.12-slim" in dockerfile


def test_dockerfile_has_healthcheck() -> None:
    dockerfile = (ROOT / "Dockerfile").read_text(encoding="utf-8")
    assert 'CMD ["orchesis", "proxy", "--config", "/app/config/orchesis.yaml"]' in dockerfile



"""Tests for Semantic Cache — fuzzy response caching with SimHash + Jaccard."""

from __future__ import annotations

import json
import threading
import time

import pytest

from orchesis.semantic_cache import (
    SemanticCache,
    SemanticCacheConfig,
)


# --- SimHash (8 tests) ---


def test_simhash_identical_text() -> None:
    """Same text → distance 0."""
    a = SemanticCache._compute_simhash("hello world")
    b = SemanticCache._compute_simhash("hello world")
    assert SemanticCache._hamming_distance(a, b) == 0


def test_simhash_similar_text() -> None:
    """Similar texts → smaller distance than unrelated."""
    a = SemanticCache._compute_simhash("the quick brown fox jumps over the lazy dog")
    b = SemanticCache._compute_simhash("the quick brown fox runs over the lazy dog")
    c = SemanticCache._compute_simhash("quantum physics and relativity theory")
    d_ab = SemanticCache._hamming_distance(a, b)
    d_ac = SemanticCache._hamming_distance(a, c)
    assert d_ab < d_ac


def test_simhash_different_text() -> None:
    """Unrelated text → large distance (>20)."""
    a = SemanticCache._compute_simhash("the quick brown fox")
    b = SemanticCache._compute_simhash("quantum physics and relativity theory")
    assert SemanticCache._hamming_distance(a, b) > 20


def test_simhash_word_order() -> None:
    """'quick brown fox' vs 'brown quick fox' → both produce valid hashes."""
    a = SemanticCache._compute_simhash("quick brown fox")
    b = SemanticCache._compute_simhash("brown quick fox")
    assert isinstance(a, int) and isinstance(b, int)
    assert SemanticCache._hamming_distance(a, a) == 0


def test_simhash_deterministic() -> None:
    """Same input → same hash always."""
    text = "deterministic test"
    a = SemanticCache._compute_simhash(text)
    b = SemanticCache._compute_simhash(text)
    assert a == b


def test_simhash_empty_text() -> None:
    """Empty → deterministic value."""
    a = SemanticCache._compute_simhash("")
    b = SemanticCache._compute_simhash("")
    assert a == b == 0


def test_simhash_long_text() -> None:
    """10K chars → works, no crash."""
    text = "word " * 2000
    h = SemanticCache._compute_simhash(text)
    assert isinstance(h, int)


def test_simhash_unicode() -> None:
    """Handles non-ASCII."""
    a = SemanticCache._compute_simhash("café résumé naïve")
    b = SemanticCache._compute_simhash("café résumé naïve")
    assert a == b


# --- Trigrams + Jaccard (7 tests) ---


def test_trigrams_basic() -> None:
    """'hello' → {'hel', 'ell', 'llo'}."""
    t = SemanticCache._compute_trigrams("hello")
    assert "hel" in t
    assert "ell" in t
    assert "llo" in t


def test_trigrams_normalized() -> None:
    """Uppercase → lowered, whitespace collapsed."""
    t = SemanticCache._compute_trigrams("HELLO   WORLD")
    assert "hel" in t or "ell" in t


def test_jaccard_identical() -> None:
    """Same set → 1.0."""
    s = frozenset(["a", "b", "c"])
    assert SemanticCache._jaccard_similarity(s, s) == 1.0


def test_jaccard_disjoint() -> None:
    """No overlap → 0.0."""
    a = frozenset(["x", "y", "z"])
    b = frozenset(["p", "q", "r"])
    assert SemanticCache._jaccard_similarity(a, b) == 0.0


def test_jaccard_partial() -> None:
    """50% overlap → ~0.33-0.5."""
    a = frozenset(["a", "b", "c"])
    b = frozenset(["b", "c", "d"])
    j = SemanticCache._jaccard_similarity(a, b)
    assert 0.2 <= j <= 0.6


def test_jaccard_empty_sets() -> None:
    """Both empty → 1.0, one empty → 0.0."""
    assert SemanticCache._jaccard_similarity(frozenset(), frozenset()) == 1.0
    assert SemanticCache._jaccard_similarity(frozenset(["a"]), frozenset()) == 0.0


def test_jaccard_symmetric() -> None:
    """J(A,B) == J(B,A)."""
    a = frozenset(["a", "b"])
    b = frozenset(["b", "c"])
    assert SemanticCache._jaccard_similarity(a, b) == SemanticCache._jaccard_similarity(b, a)


# --- Structural Key (5 tests) ---


def test_structural_key_same_structure() -> None:
    """Same model+roles+tools → same key."""
    msgs = [{"role": "user", "content": "hi"}, {"role": "assistant", "content": "hello"}]
    a = SemanticCache._compute_structural_key("gpt-4", msgs, ["read_file"])
    b = SemanticCache._compute_structural_key("gpt-4", msgs, ["read_file"])
    assert a == b


def test_structural_key_different_model() -> None:
    """Different model → different key."""
    msgs = [{"role": "user", "content": "hi"}]
    a = SemanticCache._compute_structural_key("gpt-4", msgs, None)
    b = SemanticCache._compute_structural_key("claude-3", msgs, None)
    assert a != b


def test_structural_key_different_roles() -> None:
    """Different message sequence → different key."""
    msgs1 = [{"role": "user", "content": "a"}]
    msgs2 = [{"role": "user", "content": "a"}, {"role": "assistant", "content": "b"}]
    a = SemanticCache._compute_structural_key("m", msgs1, None)
    b = SemanticCache._compute_structural_key("m", msgs2, None)
    assert a != b


def test_structural_key_different_tools() -> None:
    """Different tools → different key."""
    msgs = [{"role": "user", "content": "hi"}]
    a = SemanticCache._compute_structural_key("m", msgs, ["read_file"])
    b = SemanticCache._compute_structural_key("m", msgs, ["write_file"])
    assert a != b


def test_structural_key_no_tools() -> None:
    """tools=None handled."""
    msgs = [{"role": "user", "content": "hi"}]
    key = SemanticCache._compute_structural_key("m", msgs, None)
    assert ":" in key


# --- Cache Lookup (10 tests) ---


def test_exact_hit() -> None:
    """Identical request → hit, match_type='exact'."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "summarize this document please"}]
    body = b'{"content":[{"type":"text","text":"Summary here"}]}'
    cache.store(msgs, "gpt-4", None, body, tokens=100, cost_usd=0.001)
    result = cache.lookup(msgs, "gpt-4", None)
    assert result.hit is True
    assert result.match_type == "exact"
    assert result.response_body == body


def test_semantic_hit() -> None:
    """Similar request → hit, match_type='semantic'."""
    cache = SemanticCache(
        SemanticCacheConfig(enabled=True, simhash_threshold=15, jaccard_threshold=0.35)
    )
    msgs1 = [{"role": "user", "content": "summarize this document for me please"}]
    msgs2 = [{"role": "user", "content": "please summarize this document for me"}]
    body = b'{"content":[{"type":"text","text":"Summary"}]}'
    cache.store(msgs1, "gpt-4", None, body, tokens=50, cost_usd=0.0005)
    result = cache.lookup(msgs2, "gpt-4", None)
    assert result.hit is True
    assert result.match_type == "semantic"
    assert result.similarity >= 0.3


def test_miss_different_content() -> None:
    """Unrelated → miss."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs1 = [{"role": "user", "content": "what is the capital of France"}]
    msgs2 = [{"role": "user", "content": "write a python function to sort a list"}]
    body = b'{"content":[{"type":"text","text":"Paris"}]}'
    cache.store(msgs1, "gpt-4", None, body)
    result = cache.lookup(msgs2, "gpt-4", None)
    assert result.hit is False


def test_miss_different_structure() -> None:
    """Same text, different model → miss."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "hello world" * 5}]
    body = b'{"content":[{"type":"text","text":"Hi"}]}'
    cache.store(msgs, "gpt-4", None, body)
    result = cache.lookup(msgs, "claude-3", None)
    assert result.hit is False


def test_semantic_hit_word_order() -> None:
    """'summarize doc' vs 'doc summarize' → hit."""
    cache = SemanticCache(
        SemanticCacheConfig(enabled=True, simhash_threshold=12, jaccard_threshold=0.4)
    )
    msgs1 = [{"role": "user", "content": "summarize the document for me"}]
    msgs2 = [{"role": "user", "content": "the document summarize for me"}]
    body = b'{"content":[{"type":"text","text":"Summary"}]}'
    cache.store(msgs1, "gpt-4", None, body)
    result = cache.lookup(msgs2, "gpt-4", None)
    assert result.hit is True or result.similarity < 0.4


def test_semantic_hit_synonym() -> None:
    """'explain X' vs 'describe X' → close."""
    cache = SemanticCache(
        SemanticCacheConfig(enabled=True, simhash_threshold=10, jaccard_threshold=0.35)
    )
    msgs1 = [{"role": "user", "content": "explain how photosynthesis works in plants"}]
    msgs2 = [{"role": "user", "content": "describe how photosynthesis works in plants"}]
    body = b'{"content":[{"type":"text","text":"Explanation"}]}'
    cache.store(msgs1, "gpt-4", None, body)
    result = cache.lookup(msgs2, "gpt-4", None)
    assert result.hit is True or result.similarity >= 0.3


def test_ttl_expiration() -> None:
    """Entry expires after TTL."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, ttl_seconds=0.1))
    msgs = [{"role": "user", "content": "hello world " * 5}]
    body = b'{"ok":true}'
    cache.store(msgs, "gpt-4", None, body)
    assert cache.lookup(msgs, "gpt-4", None).hit is True
    time.sleep(0.2)
    assert cache.lookup(msgs, "gpt-4", None).hit is False


def test_lru_eviction() -> None:
    """Oldest evicted at max_entries."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, max_entries=3))
    for i in range(5):
        msgs = [{"role": "user", "content": f"unique content {i} " * 5}]
        cache.store(msgs, "gpt-4", None, b"{}")
    assert len(cache._entries) <= 3


def test_similarity_score_returned() -> None:
    """Jaccard score in result."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "test content " * 5}]
    cache.store(msgs, "gpt-4", None, b"{}")
    result = cache.lookup(msgs, "gpt-4", None)
    assert result.similarity == 1.0


def test_max_content_length_skip() -> None:
    """Too long → not cached."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, max_content_length=100))
    msgs = [{"role": "user", "content": "x" * 200}]
    ok = cache.store(msgs, "gpt-4", None, b"{}")
    assert ok is False


# --- Cache Store (6 tests) ---


def test_store_basic() -> None:
    """Stored and retrievable."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "hello " * 10}]
    body = b'{"result":"ok"}'
    assert cache.store(msgs, "gpt-4", None, body) is True
    result = cache.lookup(msgs, "gpt-4", None)
    assert result.hit is True
    assert result.response_body == body


def test_store_too_short() -> None:
    """Below min_content_length → not stored."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, min_content_length=50))
    msgs = [{"role": "user", "content": "hi"}]
    assert cache.store(msgs, "gpt-4", None, b"{}") is False


def test_store_too_long() -> None:
    """Above max_content_length → not stored."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, max_content_length=50))
    msgs = [{"role": "user", "content": "x" * 100}]
    assert cache.store(msgs, "gpt-4", None, b"{}") is False


def test_store_tool_call_excluded() -> None:
    """tool_use response not cached."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, exclude_tool_calls=True))
    msgs = [{"role": "user", "content": "read file " * 5}]
    body = b'{"content":[{"type":"message","content":[{"type":"tool_use","name":"read"}]}]}'
    assert cache.store(msgs, "gpt-4", None, body) is False


def test_store_model_filter() -> None:
    """Only cacheable_models stored."""
    cache = SemanticCache(
        SemanticCacheConfig(enabled=True, cacheable_models=["gpt-4"])
    )
    msgs = [{"role": "user", "content": "hello " * 10}]
    assert cache.store(msgs, "claude-3", None, b"{}") is False
    assert cache.store(msgs, "gpt-4", None, b"{}") is True


def test_store_updates_simhash_index() -> None:
    """Index updated after store."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "content " * 10}]
    cache.store(msgs, "gpt-4", None, b"{}")
    assert len(cache._simhash_index) >= 1


# --- Stats (5 tests) ---


def test_stats_hit_rate() -> None:
    """Correct hit rate calculation."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "test " * 10}]
    cache.store(msgs, "gpt-4", None, b"{}")
    cache.lookup(msgs, "gpt-4", None)
    cache.lookup([{"role": "user", "content": "different " * 10}], "gpt-4", None)
    stats = cache.get_stats()
    assert stats["lookups"] == 2
    assert stats["hit_rate_percent"] >= 0


def test_stats_tokens_saved() -> None:
    """Accumulated tokens."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, track_savings=True))
    msgs = [{"role": "user", "content": "x" * 50}]
    cache.store(msgs, "gpt-4", None, b"{}", tokens=100, cost_usd=0.001)
    cache.lookup(msgs, "gpt-4", None)
    stats = cache.get_stats()
    assert stats["total_tokens_saved"] >= 0


def test_stats_cost_saved() -> None:
    """Accumulated cost."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, track_savings=True))
    msgs = [{"role": "user", "content": "y" * 50}]
    cache.store(msgs, "gpt-4", None, b"{}", tokens=50, cost_usd=0.0005)
    cache.lookup(msgs, "gpt-4", None)
    stats = cache.get_stats()
    assert "total_cost_saved_usd" in stats


def test_stats_evictions_counted() -> None:
    """Eviction counter."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, max_entries=2))
    for i in range(5):
        msgs = [{"role": "user", "content": f"evict {i} " * 5}]
        cache.store(msgs, "gpt-4", None, b"{}")
    stats = cache.get_stats()
    assert stats["evictions"] >= 3


def test_stats_empty_cache() -> None:
    """Zero values when empty."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    stats = cache.get_stats()
    assert stats["entries"] == 0
    assert stats["lookups"] == 0


# --- Integration (4 tests) ---


def test_thread_safe_concurrent() -> None:
    """10 threads store+lookup."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))

    def run(i: int) -> None:
        msgs = [{"role": "user", "content": f"thread {i} content " * 5}]
        cache.store(msgs, "gpt-4", None, b"{}")
        cache.lookup(msgs, "gpt-4", None)

    threads = [threading.Thread(target=run, args=(i,)) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert cache.get_stats()["entries"] >= 1


def test_clear_cache() -> None:
    """clear() empties everything."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "clear test " * 5}]
    cache.store(msgs, "gpt-4", None, b"{}")
    cache.clear()
    assert len(cache._entries) == 0
    assert cache.lookup(msgs, "gpt-4", None).hit is False


def test_invalidate_entry() -> None:
    """Specific entry removed."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "invalidate " * 5}]
    cache.store(msgs, "gpt-4", None, b"{}")
    result = cache.lookup(msgs, "gpt-4", None)
    assert result.hit and result.cache_key
    cache.invalidate(result.cache_key)
    assert cache.lookup(msgs, "gpt-4", None).hit is False


def test_cache_doesnt_grow_unbounded() -> None:
    """Stays within max_entries."""
    cache = SemanticCache(SemanticCacheConfig(enabled=True, max_entries=10))
    for i in range(100):
        msgs = [{"role": "user", "content": f"bounded {i} " * 5}]
        cache.store(msgs, "gpt-4", None, b"{}")
    assert len(cache._entries) <= 10


# --- Proxy Integration (5 tests) ---


def test_proxy_cache_hit_skips_upstream() -> None:
    """Cached response returned."""
    from orchesis.proxy import LLMHTTPProxy, _RequestContext

    class FakeHandler:
        path = "/v1/messages"
        headers = {}

    proxy = LLMHTTPProxy(policy_path=None)
    cfg = SemanticCacheConfig(enabled=True)
    proxy._semantic_cache = SemanticCache(cfg)
    msgs = [{"role": "user", "content": "proxy cache test " * 5}]
    body = b'{"content":[{"type":"text","text":"cached"}],"usage":{"input_tokens":0,"output_tokens":0}}'
    proxy._semantic_cache.store(msgs, "gpt-4", None, body)
    ctx = _RequestContext(
        handler=FakeHandler(),
        body={"messages": msgs, "model": "gpt-4"},
    )
    ctx.parsed_req = type("Parsed", (), {"tool_calls": [], "content_text": "test", "provider": "anthropic"})()
    ok = proxy._phase_upstream(ctx)
    assert ok is True
    assert ctx.from_semantic_cache is True
    assert ctx.resp_body == body


def test_proxy_cache_headers() -> None:
    """X-Orchesis-Cache, X-Orchesis-Cache-Similarity."""
    from orchesis.proxy import LLMHTTPProxy, _RequestContext

    class FakeHandler:
        path = "/v1/messages"
        headers = {}

    proxy = LLMHTTPProxy(policy_path=None)
    proxy._semantic_cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "headers test " * 5}]
    proxy._semantic_cache.store(msgs, "gpt-4", None, b"{}")
    ctx = _RequestContext(handler=FakeHandler(), body={"messages": msgs, "model": "gpt-4"})
    ctx.parsed_req = type("Parsed", (), {"tool_calls": [], "content_text": "test", "provider": "anthropic"})()
    proxy._phase_upstream(ctx)
    assert ctx.from_semantic_cache
    assert "X-Orchesis-Cache" in ctx.session_headers


def test_proxy_stores_after_upstream() -> None:
    """Successful response cached."""
    from orchesis.proxy import LLMHTTPProxy, _RequestContext

    class FakeHandler:
        path = "/v1/messages"
        headers = {}

    proxy = LLMHTTPProxy(policy_path=None)
    proxy._semantic_cache = SemanticCache(SemanticCacheConfig(enabled=True))
    msgs = [{"role": "user", "content": "store after " * 5}]
    ctx = _RequestContext(
        handler=FakeHandler(),
        body={"messages": msgs, "model": "gpt-4"},
    )
    ctx.parsed_req = type("Parsed", (), {"tool_calls": [], "content_text": "test", "provider": "anthropic"})()
    ctx.provider = "anthropic"
    ctx.resp_status = 200
    ctx.resp_body = b'{"content":[{"type":"text","text":"ok"}],"usage":{"input_tokens":10,"output_tokens":5}}'
    ctx.from_semantic_cache = False
    ctx.is_streaming = False
    ctx.proc_result = {"cost": 0.001, "allowed": True}
    proxy._phase_post_upstream(ctx)
    stats = proxy._semantic_cache.get_stats()
    assert stats["entries"] >= 1


def test_proxy_cache_stats() -> None:
    """Stats includes semantic_cache."""
    from orchesis.proxy import LLMHTTPProxy

    proxy = LLMHTTPProxy(policy_path=None)
    proxy._semantic_cache = SemanticCache(SemanticCacheConfig(enabled=True))
    stats = proxy.stats
    assert "semantic_cache" in stats
    assert "hit_rate_percent" in stats["semantic_cache"]


def test_config_normalization() -> None:
    """semantic_cache config validated."""
    import tempfile
    from pathlib import Path

    from orchesis.config import load_policy

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        f.write(
            """
semantic_cache:
  enabled: true
  max_entries: 2000
  ttl_seconds: 600
  simhash_threshold: 8
  jaccard_threshold: 0.6
  min_content_length: 20
  max_content_length: 50000
  cacheable_models: []
  exclude_tool_calls: true
  track_savings: true
"""
        )
        path = f.name
    try:
        policy = load_policy(path)
        assert "semantic_cache" in policy
        sc = policy["semantic_cache"]
        assert isinstance(sc, dict)
        assert sc.get("enabled") is True
    finally:
        import os

        os.unlink(path)



"""Parse LLM API requests/responses to extract tool calls and usage."""

from __future__ import annotations

from dataclasses import dataclass, field
import json
from typing import Any


@dataclass
class ToolCall:
    """Represents an extracted tool call."""

    name: str
    params: dict[str, Any] = field(default_factory=dict)
    call_id: str = ""


@dataclass
class ParsedRequest:
    """Parsed incoming LLM API request."""

    provider: str
    model: str = ""
    messages: list[dict[str, Any]] = field(default_factory=list)
    tool_calls: list[ToolCall] = field(default_factory=list)
    tool_definitions: list[dict[str, Any]] = field(default_factory=list)
    raw_body: dict[str, Any] = field(default_factory=dict)
    content_text: str = ""


@dataclass
class ParsedResponse:
    """Parsed LLM API response payload."""

    provider: str
    model: str = ""
    tool_calls: list[ToolCall] = field(default_factory=list)
    content_text: str = ""
    input_tokens: int = 0
    output_tokens: int = 0
    raw_body: dict[str, Any] = field(default_factory=dict)
    stop_reason: str = ""


def parse_request(body: dict[str, Any], path: str) -> ParsedRequest:
    """Parse incoming request body from Anthropic/OpenAI-compatible APIs."""
    safe_path = path.lower()
    if "/messages" in safe_path:
        return _parse_anthropic_request(body)
    if "/chat/completions" in safe_path:
        return _parse_openai_request(body)
    if "max_tokens" in body and "messages" in body:
        return _parse_anthropic_request(body)
    return _parse_openai_request(body)


def parse_response(body: dict[str, Any], provider: str) -> ParsedResponse:
    """Parse response body from Anthropic/OpenAI-compatible APIs."""
    safe_provider = provider.lower().strip()
    if safe_provider == "anthropic":
        return _parse_anthropic_response(body)
    return _parse_openai_response(body)


def _parse_anthropic_request(body: dict[str, Any]) -> ParsedRequest:
    result = ParsedRequest(provider="anthropic", raw_body=body)
    result.model = str(body.get("model", ""))
    messages = body.get("messages", [])
    result.messages = messages if isinstance(messages, list) else []
    tools = body.get("tools", [])
    result.tool_definitions = tools if isinstance(tools, list) else []

    text_parts: list[str] = []
    for msg in result.messages:
        if not isinstance(msg, dict):
            continue
        content = msg.get("content", "")
        if isinstance(content, str):
            if content:
                text_parts.append(content)
            continue
        if not isinstance(content, list):
            continue
        for block in content:
            if not isinstance(block, dict):
                continue
            block_type = str(block.get("type", ""))
            if block_type == "tool_use":
                name = block.get("name", "")
                call_input = block.get("input", {})
                result.tool_calls.append(
                    ToolCall(
                        name=str(name) if isinstance(name, str) else "",
                        params=call_input if isinstance(call_input, dict) else {},
                        call_id=str(block.get("id", "")),
                    )
                )
            elif block_type == "text":
                text = block.get("text", "")
                if isinstance(text, str) and text:
                    text_parts.append(text)
            elif block_type == "tool_result":
                content_val = block.get("content", "")
                if isinstance(content_val, str) and content_val:
                    text_parts.append(content_val)
                elif isinstance(content_val, list):
                    for sub in content_val:
                        if isinstance(sub, dict) and sub.get("type") == "text":
                            text = sub.get("text", "")
                            if isinstance(text, str) and text:
                                text_parts.append(text)

    result.content_text = "\n".join(text_parts)
    return result


def _parse_openai_request(body: dict[str, Any]) -> ParsedRequest:
    result = ParsedRequest(provider="openai", raw_body=body)
    result.model = str(body.get("model", ""))
    messages = body.get("messages", [])
    result.messages = messages if isinstance(messages, list) else []
    tools = body.get("tools", [])
    result.tool_definitions = tools if isinstance(tools, list) else []

    text_parts: list[str] = []
    for msg in result.messages:
        if not isinstance(msg, dict):
            continue
        content = msg.get("content", "")
        if isinstance(content, str) and content:
            text_parts.append(content)
        tool_calls = msg.get("tool_calls", [])
        if not isinstance(tool_calls, list):
            continue
        for tc in tool_calls:
            if not isinstance(tc, dict):
                continue
            func = tc.get("function", {})
            if not isinstance(func, dict):
                continue
            name = func.get("name", "")
            arguments = func.get("arguments", "{}")
            params: dict[str, Any]
            try:
                parsed = json.loads(arguments) if isinstance(arguments, str) else arguments
            except (json.JSONDecodeError, TypeError):
                parsed = {}
            params = parsed if isinstance(parsed, dict) else {}
            result.tool_calls.append(
                ToolCall(
                    name=str(name) if isinstance(name, str) else "",
                    params=params,
                    call_id=str(tc.get("id", "")),
                )
            )

    result.content_text = "\n".join(text_parts)
    return result


def _parse_anthropic_response(body: dict[str, Any]) -> ParsedResponse:
    result = ParsedResponse(provider="anthropic", raw_body=body)
    result.model = str(body.get("model", ""))
    result.stop_reason = str(body.get("stop_reason", ""))
    usage = body.get("usage", {})
    if isinstance(usage, dict):
        result.input_tokens = int(usage.get("input_tokens", 0) or 0)
        result.output_tokens = int(usage.get("output_tokens", 0) or 0)

    content = body.get("content", [])
    text_parts: list[str] = []
    if isinstance(content, list):
        for block in content:
            if not isinstance(block, dict):
                continue
            block_type = str(block.get("type", ""))
            if block_type == "text":
                text = block.get("text", "")
                if isinstance(text, str) and text:
                    text_parts.append(text)
            elif block_type == "tool_use":
                name = block.get("name", "")
                call_input = block.get("input", {})
                result.tool_calls.append(
                    ToolCall(
                        name=str(name) if isinstance(name, str) else "",
                        params=call_input if isinstance(call_input, dict) else {},
                        call_id=str(block.get("id", "")),
                    )
                )

    result.content_text = "\n".join(text_parts)
    return result


def _parse_openai_response(body: dict[str, Any]) -> ParsedResponse:
    result = ParsedResponse(provider="openai", raw_body=body)
    result.model = str(body.get("model", ""))
    usage = body.get("usage", {})
    if isinstance(usage, dict):
        result.input_tokens = int(usage.get("prompt_tokens", 0) or 0)
        result.output_tokens = int(usage.get("completion_tokens", 0) or 0)

    text_parts: list[str] = []
    choices = body.get("choices", [])
    if isinstance(choices, list):
        for choice in choices:
            if not isinstance(choice, dict):
                continue
            finish_reason = choice.get("finish_reason")
            if isinstance(finish_reason, str):
                result.stop_reason = finish_reason
            msg = choice.get("message", {})
            if not isinstance(msg, dict):
                continue
            content = msg.get("content", "")
            if isinstance(content, str) and content:
                text_parts.append(content)
            tool_calls = msg.get("tool_calls", [])
            if not isinstance(tool_calls, list):
                continue
            for tc in tool_calls:
                if not isinstance(tc, dict):
                    continue
                func = tc.get("function", {})
                if not isinstance(func, dict):
                    continue
                name = func.get("name", "")
                arguments = func.get("arguments", "{}")
                try:
                    parsed = json.loads(arguments) if isinstance(arguments, str) else arguments
                except (json.JSONDecodeError, TypeError):
                    parsed = {}
                params = parsed if isinstance(parsed, dict) else {}
                result.tool_calls.append(
                    ToolCall(
                        name=str(name) if isinstance(name, str) else "",
                        params=params,
                        call_id=str(tc.get("id", "")),
                    )
                )

    result.content_text = "\n".join(text_parts)
    return result




"""External notification integrations for Orchesis."""

from orchesis.integrations.alert_manager import AlertManager
from orchesis.integrations.base import AlertEvent, BaseIntegration
from orchesis.integrations.discord import DiscordIntegration
from orchesis.integrations.forensics_emitter import ForensicsEmitter
from orchesis.integrations.github import GitHubIntegration
from orchesis.integrations.slack import SlackEmitter, SlackIntegration, SlackNotifier
from orchesis.integrations.telegram import TelegramBot, TelegramEmitter, TelegramIntegration, TelegramNotifier
from orchesis.integrations.webhook import WebhookIntegration

__all__ = [
    "AlertEvent",
    "BaseIntegration",
    "AlertManager",
    "ForensicsEmitter",
    "DiscordIntegration",
    "GitHubIntegration",
    "SlackIntegration",
    "SlackNotifier",
    "SlackEmitter",
    "TelegramBot",
    "TelegramIntegration",
    "TelegramNotifier",
    "TelegramEmitter",
    "WebhookIntegration",
]



"""Local-only network exposure scanner."""

from __future__ import annotations

import os
import platform
import socket
import stat
from pathlib import Path
from typing import Any

from orchesis.contrib.secret_scanner import SecretScanner


class NetworkExposureScanner:
    """Scan local environment for network exposure risks."""

    def __init__(self):
        self._secret_scanner = SecretScanner()

    def scan_all(self) -> list[dict[str, Any]]:
        findings: list[dict[str, Any]] = []
        for method in (
            self.check_open_ports,
            self.check_firewall_status,
            self.check_config_file_permissions,
            self.check_env_files,
            self.check_known_agent_configs,
        ):
            try:
                findings.extend(method())
            except Exception:
                continue
        return findings

    def check_open_ports(self) -> list[dict[str, Any]]:
        findings: list[dict[str, Any]] = []
        ports = [3000, 3001, 8080, 8081, 18789, 9090, 5432, 3306, 27017]
        for port in ports:
            try:
                localhost_open = self._can_connect("127.0.0.1", port)
                zero_open = self._can_connect("0.0.0.0", port)
                if localhost_open and zero_open:
                    findings.append(
                        self._finding(
                            "open_port",
                            "critical",
                            f"Port {port} appears externally accessible",
                            f"0.0.0.0:{port}",
                            "Bind service to 127.0.0.1 or enforce authentication.",
                        )
                    )
                elif localhost_open:
                    findings.append(
                        self._finding(
                            "localhost_ws",
                            "info",
                            f"Port {port} listening on localhost",
                            f"127.0.0.1:{port}",
                            "Localhost services can still be targeted via browser-based attacks.",
                        )
                    )
            except Exception:
                continue
        return findings

    def check_firewall_status(self) -> list[dict[str, Any]]:
        findings: list[dict[str, Any]] = []
        try:
            os_name = platform.system().lower()
            if os_name == "windows":
                return [self._finding("firewall_status", "info", "Firewall status check limited on Windows", "netsh required", "Verify Windows Defender Firewall is enabled.")]
            if os_name == "darwin":
                if not Path("/etc/pf.conf").exists():
                    findings.append(
                        self._finding(
                            "no_firewall",
                            "medium",
                            "No pf firewall configuration detected",
                            "/etc/pf.conf missing",
                            "Enable pf or host firewall controls.",
                        )
                    )
            else:
                if not any(Path(path).exists() for path in ("/sbin/iptables", "/usr/sbin/nft")):
                    findings.append(
                        self._finding(
                            "no_firewall",
                            "medium",
                            "No firewall tooling detected",
                            "iptables/nft not found",
                            "Enable ufw/iptables/nftables.",
                        )
                    )
        except Exception:
            return []
        return findings

    def check_config_file_permissions(self) -> list[dict[str, Any]]:
        findings: list[dict[str, Any]] = []
        candidates = [
            Path.home() / ".clawdbot",
            Path.home() / ".config" / "claude",
            Path.home() / ".cursor" / "mcp.json",
            Path(".vscode") / "mcp.json",
            Path("policy.yaml"),
            Path(".env"),
            Path(".env.local"),
            Path(".env.production"),
            Path("openclaw.json"),
            Path("device.json"),
            Path("soul.md"),
        ]
        if os.name == "nt":
            return findings
        for path in candidates:
            try:
                if not path.exists():
                    continue
                st_mode = path.stat().st_mode
                if bool(st_mode & stat.S_IROTH):
                    findings.append(
                        self._finding(
                            "config_perms",
                            "high",
                            f"{path} is world-readable",
                            oct(st_mode & 0o777),
                            f"chmod 600 {path}",
                        )
                    )
                elif bool(st_mode & stat.S_IRGRP):
                    findings.append(
                        self._finding(
                            "config_perms",
                            "medium",
                            f"{path} is group-readable",
                            oct(st_mode & 0o777),
                            f"chmod 640 or stricter for {path}",
                        )
                    )
            except Exception:
                continue
        return findings

    def check_env_files(self) -> list[dict[str, Any]]:
        findings: list[dict[str, Any]] = []
        env_paths = [
            Path(".env"),
            Path(".env.local"),
            Path(".env.production"),
            Path(".env.development"),
        ]
        for env_path in env_paths:
            try:
                if not env_path.exists() or not env_path.is_file():
                    continue
                content = env_path.read_text(encoding="utf-8")
                secrets = self._secret_scanner.scan_text(content)
                if not secrets:
                    continue
                severity = "high"
                if os.name != "nt":
                    mode = env_path.stat().st_mode
                    if bool(mode & stat.S_IROTH):
                        severity = "critical"
                findings.append(
                    self._finding(
                        "env_secrets",
                        severity,
                        f"{env_path} contains {len(secrets)} potential secret(s)",
                        ", ".join(str(item["pattern"]) for item in secrets[:3]),
                        "Move secrets to secure vault/environment and restrict file permissions.",
                    )
                )
            except Exception:
                continue
        return findings

    def check_known_agent_configs(self) -> list[dict[str, Any]]:
        findings: list[dict[str, Any]] = []
        openclaw_path = Path("openclaw.json")
        try:
            if openclaw_path.exists():
                content = openclaw_path.read_text(encoding="utf-8")
                if "auth" not in content.lower():
                    findings.append(
                        self._finding(
                            "known_agent_config",
                            "high",
                            "openclaw.json appears to miss authentication config",
                            str(openclaw_path),
                            "Configure gateway authentication mode and non-default token.",
                        )
                    )
        except Exception:
            pass

        soul_path = Path("soul.md")
        try:
            if soul_path.exists():
                secrets = self._secret_scanner.scan_text(soul_path.read_text(encoding="utf-8"))
                if secrets:
                    findings.append(
                        self._finding(
                            "known_agent_config",
                            "high",
                            "soul.md may contain credential-like secrets",
                            str(soul_path),
                            "Remove secrets from soul.md and rotate credentials.",
                        )
                    )
        except Exception:
            pass
        return findings

    def _can_connect(self, host: str, port: int) -> bool:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(0.1)
        try:
            return sock.connect_ex((host, int(port))) == 0
        finally:
            sock.close()

    def _finding(
        self,
        check: str,
        severity: str,
        description: str,
        evidence: str,
        recommendation: str,
    ) -> dict[str, Any]:
        return {
            "check": check,
            "severity": severity,
            "description": description,
            "evidence": evidence,
            "recommendation": recommendation,
        }
